"""
Defines flight controller vendor, board name and MCU series depending on hwdef.dat file information.

File automatically generated by the update_flight_controller_ids.py script
Do not edit directly. ALL CHANGES WILL BE OVERWRITTEN

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

# pylint: disable=line-too-long

# Maps USB VID to vendor(s) (manufacturer) name
VID_VENDOR_DICT: dict[int, list[str]] = {
    0x26AC: ["3DR"],
    0x1209: ["ArduPilot", "AET", "Airvolute", "Aocoda-RC-H743Dual", "CSKY", "Japan Aviation Electronics Industry Ltd.", "mRo", "Horizon31", "Dheeran labs", "Qiotek", "Sierra Aerospace", "SkySakura"],
    0x2DAE: ["Hex/ProfiCNC", "CubePilot"],
    0x3162: ["Holybro"],
    0x27AC: ["Swift-Flyer", "Laser Navigation"],
    0x35A7: ["ThePeach"],
}

# Maps USB VID,PID tuple to product name(s)
VID_PID_PRODUCT_DICT: dict[tuple[int, int], list[str]] = {
    (0x26AC, 0x1124): ["CZOEMrevG"],
    (0x1209, 0x5741): ["ACNS-CM4Pilot", "ACNS-F405AIO", "AeroFox-GNSS_F9P", "AEROFOX-H7", "AeroFox-PMU", "AET-H743-Basic", "airbotf4", "AIRLink", "Airvolute-DCS2", "AnyleafH7", "Aocoda-RC-H743Dual", "AR-F407SmartBat", "ARK_CANNODE", "ARK_FPV", "ARKV6X", "AtomRCF405NAVI", "BeastF7", "BeastF7v2", "BeastH7", "BeastH7v2", "BETAFPV-F405", "BirdCANdy", "BlitzF745", "BlitzF745AIO", "BlitzH743Pro", "BlitzMiniF745", "BlitzWingH743", "BotBloxDroneNet", "C-RTK2-HP", "CarbonixF405", "CarbonixL496", "CBU-H7-LC-Stamp", "CBU-H7-Stamp", "CrazyF405", "crazyflie2", "CSKY405", "CUAV-7-Nano", "CUAV-Nora", "CUAV-Pixhack-v3", "CUAV-X7", "CUAVv5", "CUAVv5Nano", "CubeNode", "CubeRedSecondary", "CubeRedSecondary-IO", "CubeSolo", "DevEBoxH7v2", "DroneerF405", "DrotekP3Pro", "F35Lightning", "FlyingMoonF407", "FlyingMoonF427", "FlyingMoonH743", "FlywooF405HD-AIOv2", "FlywooF405Pro", "FlywooF405S-AIO", "FlywooF745", "FlywooF745Nano", "FlywooH743Pro", "fmuv2", "fmuv3", "fmuv5", "FoxeerF405v2", "FoxeerH743v1", "FreeflyRTK", "GEPRC_TAKER_H743", "GEPRCF745BTHD", "%BOARD%", "HEEWING-F405", "HEEWING-F405v2", "Here4AP", "Here4FC", "HolybroF4_PMU", "IFLIGHT_2RAW_H7", "JFB100", "JFB-110", "JHEM_JHEF405", "JHEMCU-GSF405A", "JHEMCU-GSF405A-RX2", "JHEMCU-H743HD", "KakuteF4", "KakuteF4-Wing", "KakuteF4Mini", "KakuteF7", "KakuteF7Mini", "KakuteH7", "KakuteH7-Wing", "KakuteH7Mini", "KakuteH7Mini-Nand", "KakuteH7v2", "LongBowF405WING", "LumenierLUXF765-NDAA", "luminousbee4", "luminousbee5", "MambaF405-2022", "MambaF405v2", "MambaH743v4", "MatekF405", "MatekF405-CAN", "MatekF405-STD", "MatekF405-TE", "MatekF405-Wing", "MatekF765-SE", "MatekF765-Wing", "MatekG474", "MatekH743", "MatekH7A3", "MazzyStarDrone", "MFE_POS3_CAN", "MFT-SEMA100", "MicoAir405Mini", "MicoAir405v2", "MicoAir743", "MicoAir743-AIO", "MicoAir743v2", "mindpx-v2", "mini-pix", "modalai_fc-v1", "mRo-M10095", "mRoCANPWM-M10126", "mRoControlZeroClassic", "mRoControlZeroF7", "mRoControlZeroH7", "mRoControlZeroOEMH7", "mRoCZeroOEMH7", "mRoNexus", "mRoPixracerPro", "mRoX21", "mRoX21-777", "MUPilot", "Nucleo-G491", "Nucleo-L476", "Nucleo-L496", "NucleoH743", "NucleoH755", "NxtPX4v2", "omnibusf4", "omnibusf4pro", "omnibusf4pro-one", "omnibusf4v6", "OMNIBUSF7V2", "OmnibusNanoV6", "OrqaF405Pro", "PH4-mini", "PixC4-Jetson", "PixFlamingo", "Pixhawk1", "Pixhawk1-1M", "Pixhawk4", "Pixhawk6X", "Pixhawk6X-PPPGW", "PixPilot-C3", "PixPilot-V3", "PixPilot-V6", "PixPilot-V6PRO", "Pixracer", "PixSurveyA1", "PixSurveyA1-IND", "PixSurveyA2", "QioTekAdeptF407", "QioTekZealotF427", "QioTekZealotH743", "R9Pilot", "RadiolinkPIX6", "RADIX2HD", "ReaperF745", "revo-mini", "revo-mini-i2c", "revo-mini-sd", "rFCU", "rGNSS", "SDMODELH7V1", "SDMODELH7V2", "Sierra-F405", "Sierra-F412", "Sierra-F9P", "Sierra-L431", "Sierra-PrecisionPoint", "Sierra-TrueNavIC", "Sierra-TrueNavPro", "Sierra-TrueNavPro-G4", "Sierra-TrueNorth", "Sierra-TrueSpeed", "SIYI_N7", "SkySakuraH743", "SkystarsH7HD", "skyviper-f412-rev1", "skyviper-journey", "skyviper-v2450", "sparky2", "speedybeef4", "SpeedyBeeF405AIO", "SpeedyBeeF405Mini", "SpeedyBeeF405WING", "speedybeef4v3", "speedybeef4v4", "SPRacingH7", "SPRacingH7RF", "StellarF4", "StellarF4V2", "StellarH7V2", "SuccexF4", "SULILGH7-P1-P2", "sw-boom-f407", "sw-nav-f405", "sw-spar-f407", "Swan-K1", "TBS-Colibri-F7", "TBS_LUCID_H7", "TBS_LUCID_PRO", "TMotorH743", "VUAV-V7pro", "X-MAV-AP-H743v2", "YJUAV_A6", "YJUAV_A6SE", "YJUAV_A6SE_H743", "YJUAV_A6Ultra", "ZeroOneX6", "ZeroOneX6_Air"],
    (0x2DAE, 0x1011): ["CubeBlack", "CubeGreen-solo"],
    (0x2DAE, 0x1101): ["CubeBlack+"],
    (0x2DAE, 0x1016): ["CubeOrange"],
    (0x2DAE, 0x1057): ["Joey"],
    (0x2DAE, 0x1058): ["CubeOrange+"],
    (0x2DAE, 0x1015): ["CubePurple"],
    (0x2DAE, 0x1059): ["CubeRed"],
    (0x2DAE, 0x1012): ["CubeYellow"],
    (0x3162, 0x004B): ["Durandal"],
    (0x27AC, 0x0201): ["F4BY", "F4BY_F427"],
    (0x3162, 0x004E): ["Pix32v5"],
    (0x3162, 0x0051): ["Pixhawk5X"],
    (0x3162, 0x0053): ["Pixhawk6C"],
    (0x35A7, 0x0001): ["FCC-K1"],
    (0x35A7, 0x0002): ["FCC-R1"],
    (0x27AC, 0x1151): ["VRBrain-v51"],
    (0x27AC, 0x1152): ["VRBrain-v52"],
    (0x27AC, 0x1154): ["VRBrain-v54"],
    (0x27AC, 0x1910): ["VRCore-v10"],
    (0x27AC, 0x1351): ["VRUBrain-v51"],
}

# Maps 16-bit APJ board ID to board name(s) for 246 supported boards
APJ_BOARD_ID_NAME_DICT: dict[int, list[str]] = {
    1124: ["3DRControlZeroG"],
    1115: ["ACNS-CM4Pilot"],
    1116: ["ACNS-F405AIO"],
    1109: ["AeroFox-GNSS_F9P"],
    7110: ["AEROFOX-H7"],
    1080: ["AeroFox-PMU"],
    2024: ["AET-H743-Basic"],
    128: ["airbotf4"],
    55: ["AIRLink"],
    5200: ["Airvolute-DCS2"],
    1146: ["AnyleafH7"],
    5210: ["Aocoda-RC-H743Dual"],
    1134: ["AR-F407SmartBat"],
    83: ["ARK_CANNODE"],
    59: ["ARK_FPV"],
    57: ["ARKV6X"],
    1078: ["AtomRCF405NAVI"],
    1026: ["BeastF7"],
    1057: ["BeastF7v2"],
    1025: ["BeastH7"],
    1056: ["BeastH7v2"],
    1125: ["BETAFPV-F405"],
    1044: ["BirdCANdy"],
    1164: ["BlitzF745"],
    1117: ["BlitzF745AIO"],
    1162: ["BlitzH743Pro"],
    1163: ["BlitzMiniF745"],
    1168: ["BlitzWingH743"],
    1148: ["BotBloxDroneNet"],
    1085: ["C-RTK2-HP"],
    1064: ["CarbonixF405"],
    1053: ["CarbonixL496"],
    1182: ["CBU-H7-LC-Stamp"],
    1156: ["CBU-H7-Stamp"],
    1177: ["CrazyF405"],
    12: ["crazyflie2"],
    1158: ["CSKY405"],
    7000: ["CUAV-7-Nano"],
    1009: ["CUAV-Nora"],
    9: ["CUAV-Pixhack-v3", "CubeBlack", "CubeGreen-solo", "CubePurple", "CubeSolo", "fmuv2", "fmuv3", "mRoX21", "Pixhawk1", "Pixhawk1-1M", "skyviper-f412-rev1", "skyviper-journey", "skyviper-v2450"],
    1010: ["CUAV-X7"],
    50: ["CUAVv5", "CUAVv5Nano", "fmuv5", "PH4-mini", "Pix32v5", "Pixhawk4", "Swan-K1", "TBS-Colibri-F7"],
    1003: ["CubeBlack+"],
    1079: ["CubeNode"],
    140: ["CubeOrange"],
    1033: ["CubeOrange-joey"],
    1063: ["CubeOrangePlus"],
    1069: ["CubeRedPrimary"],
    1409: ["CubeRedPrimary-PPPGW"],
    1070: ["CubeRedSecondary", "CubeRedSecondary-IO"],
    120: ["CubeYellow"],
    1061: ["DevEBoxH7v2"],
    5800: ["DroneerF405"],
    13: ["DrotekP3Pro"],
    139: ["Durandal", "NucleoH743", "NucleoH755"],
    135: ["F35Lightning"],
    20: ["F4BY"],
    1530: ["F4BY_F427"],
    1067: ["FlyingMoonF407"],
    1068: ["FlyingMoonF427"],
    1112: ["FlyingMoonH743"],
    1180: ["FlywooF405HD-AIOv2"],
    1137: ["FlywooF405Pro"],
    1099: ["FlywooF405S-AIO"],
    1027: ["FlywooF745"],
    1042: ["FlywooF745Nano"],
    1181: ["FlywooH743Pro"],
    1157: ["FoxeerF405v2"],
    1089: ["FoxeerH743v1"],
    1028: ["FreeflyRTK"],
    1502: ["GEPRC_TAKER_H743"],
    1501: ["GEPRCF745BTHD"],
    146: ["H757I_EVAL", "H757I_EVAL_intf"],
    1119: ["HEEWING-F405", "HEEWING-F405v2"],
    1043: ["Here4AP", "Here4FC"],
    5401: ["HolybroF4_PMU"],
    1173: ["IFLIGHT_2RAW_H7"],
    1084: ["JFB100"],
    1110: ["JFB110"],
    1081: ["JHEM_JHEF405"],
    1059: ["JHEMCU-GSF405A", "JHEMCU-GSF405A-RX2"],
    1411: ["JHEMCU-H743HD"],
    122: ["KakuteF4"],
    5406: ["KakuteF4-Wing"],
    1030: ["KakuteF4Mini"],
    123: ["KakuteF7"],
    145: ["KakuteF7Mini"],
    1048: ["KakuteH7", "KakuteH7v2"],
    1105: ["KakuteH7-Wing"],
    1058: ["KakuteH7Mini", "KakuteH7Mini-Nand"],
    1315: ["kha_eth"],
    1422: ["LongBowF405WING"],
    4500: ["LumenierLUXF765-NDAA"],
    11: ["luminousbee4", "Pixracer"],
    1029: ["luminousbee5"],
    1038: ["MambaF405-2022"],
    1019: ["MambaF405v2"],
    1073: ["MambaH743v4"],
    125: ["MatekF405", "MatekF405-STD"],
    1014: ["MatekF405-CAN"],
    1054: ["MatekF405-TE"],
    127: ["MatekF405-Wing"],
    143: ["MatekF765-SE", "MatekF765-Wing"],
    1170: ["MatekG474"],
    1013: ["MatekH743"],
    1149: ["MatekH7A3"],
    188: ["MazzyStarDrone"],
    6101: ["MFE_POS3_CAN"],
    2000: ["MFT-SEMA100"],
    1161: ["MicoAir405Mini"],
    1150: ["MicoAir405v2"],
    1166: ["MicoAir743"],
    1176: ["MicoAir743-AIO"],
    1179: ["MicoAir743v2"],
    88: ["mindpx-v2"],
    3: ["mini-pix"],
    41775: ["modalai_fc-v1"],
    1041: ["mRo-M10095"],
    1098: ["mRoCANPWM-M10126"],
    1022: ["mRoControlZeroClassic"],
    141: ["mRoControlZeroF7"],
    1023: ["mRoControlZeroH7"],
    1024: ["mRoControlZeroOEMH7", "mRoCZeroOEMH7"],
    1015: ["mRoNexus"],
    1017: ["mRoPixracerPro"],
    136: ["mRoX21-777"],
    1222: ["MUPilot"],
    1040: ["Nucleo-G491"],
    1051: ["Nucleo-L476"],
    1047: ["Nucleo-L496"],
    1159: ["NxtPX4v2"],
    1002: ["omnibusf4"],
    131: ["omnibusf4pro", "omnibusf4pro-one"],
    137: ["omnibusf4v6"],
    121: ["OMNIBUSF7V2"],
    133: ["OmnibusNanoV6"],
    1155: ["OrqaF405Pro"],
    1032: ["PixC4-Jetson"],
    1090: ["PixFlamingo"],
    1131: ["PixFlamingo-F767"],
    51: ["Pixhawk5X"],
    56: ["Pixhawk6C"],
    53: ["Pixhawk6X"],
    1408: ["Pixhawk6X-PPPGW"],
    1140: ["PixPilot-C3"],
    1096: ["PixPilot-V3"],
    1083: ["PixPilot-V6"],
    1160: ["PixPilot-V6PRO"],
    1076: ["PixSurveyA1"],
    1107: ["PixSurveyA1-IND"],
    1097: ["PixSurveyA2"],
    1065: ["QioTekAdeptF407"],
    1021: ["QioTekZealotF427"],
    1036: ["QioTekZealotH743"],
    1008: ["R9Pilot"],
    1410: ["RadiolinkPIX6"],
    1118: ["RADIX2HD"],
    1074: ["ReaperF745"],
    124: ["revo-mini", "revo-mini-i2c", "revo-mini-sd"],
    1102: ["rFCU"],
    1103: ["rGNSS"],
    1111: ["SDMODELH7V1"],
    1167: ["SDMODELH7V2"],
    1052: ["Sierra-F405"],
    1055: ["Sierra-F412"],
    1034: ["Sierra-F9P"],
    1050: ["Sierra-L431"],
    1095: ["Sierra-PrecisionPoint"],
    5302: ["Sierra-TrueNavIC"],
    1091: ["Sierra-TrueNavPro"],
    5301: ["Sierra-TrueNavPro-G4"],
    1093: ["Sierra-TrueNorth"],
    1094: ["Sierra-TrueSpeed"],
    1123: ["SIYI_N7"],
    2714: ["SkySakuraH743"],
    1075: ["SkystarsH7HD"],
    130: ["sparky2"],
    134: ["speedybeef4"],
    5271: ["SpeedyBeeF405AIO"],
    1135: ["SpeedyBeeF405Mini"],
    1106: ["SpeedyBeeF405WING"],
    1082: ["speedybeef4v3"],
    1136: ["speedybeef4v4"],
    1060: ["SPRacingH7"],
    1108: ["SPRacingH7RF"],
    1500: ["StellarF4"],
    1504: ["StellarF4V2"],
    1503: ["StellarH7V2"],
    1011: ["SuccexF4"],
    2005: ["SULILGH7-P1-P2"],
    6001: ["sw-boom-f407"],
    6002: ["sw-nav-f405"],
    6000: ["sw-spar-f407"],
    5250: ["TBS_LUCID_H7"],
    5251: ["TBS_LUCID_PRO"],
    212: ["thepeach-k1"],
    213: ["thepeach-r1"],
    1138: ["TMotorH743"],
    1151: ["VRBrain-v51"],
    1152: ["VRBrain-v52"],
    1154: ["VRBrain-v54"],
    1910: ["VRCore-v10"],
    1351: ["VRUBrain-v51"],
    7100: ["VUAV-V7pro"],
    1174: ["X-MAV-AP-H743v2"],
    1113: ["YJUAV_A6"],
    1127: ["YJUAV_A6SE"],
    1141: ["YJUAV_A6SE_H743"],
    1144: ["YJUAV_A6Ultra"],
    5600: ["ZeroOneX6", "ZeroOneX6_Air"],
}

# Maps 16-bit APJ board ID to board vendor for 246 supported boards
APJ_BOARD_ID_VENDOR_DICT: dict[int, list[str]] = {
    1124: ["3DR"],  # 3DRControlZeroG
    1115: ["ArduPilot"],  # ACNS-CM4Pilot
    1116: ["ArduPilot"],  # ACNS-F405AIO
    1109: ["ArduPilot"],  # AeroFox-GNSS_F9P
    7110: ["ArduPilot"],  # AEROFOX-H7
    1080: ["ArduPilot"],  # AeroFox-PMU
    2024: ["AET"],  # AET-H743-Basic
    128: ["ArduPilot"],  # airbotf4
    55: ["ArduPilot"],  # AIRLink
    5200: ["Airvolute"],  # Airvolute-DCS2
    1146: ["ArduPilot"],  # AnyleafH7
    5210: ["Aocoda-RC-H743Dual"],  # Aocoda-RC-H743Dual
    1134: ["ArduPilot"],  # AR-F407SmartBat
    83: ["ArduPilot"],  # ARK_CANNODE
    59: ["ArduPilot"],  # ARK_FPV
    57: ["ArduPilot"],  # ARKV6X
    1078: ["ArduPilot"],  # AtomRCF405NAVI
    1026: ["ArduPilot"],  # BeastF7
    1057: ["ArduPilot"],  # BeastF7v2
    1025: ["ArduPilot"],  # BeastH7
    1056: ["ArduPilot"],  # BeastH7v2
    1125: ["ArduPilot"],  # BETAFPV-F405
    1044: ["ArduPilot"],  # BirdCANdy
    1164: ["ArduPilot"],  # BlitzF745
    1117: ["ArduPilot"],  # BlitzF745AIO
    1162: ["ArduPilot"],  # BlitzH743Pro
    1163: ["ArduPilot"],  # BlitzMiniF745
    1168: ["ArduPilot"],  # BlitzWingH743
    1148: ["ArduPilot"],  # BotBloxDroneNet
    1085: ["ArduPilot"],  # C-RTK2-HP
    1064: ["ArduPilot"],  # CarbonixF405
    1053: ["ArduPilot"],  # CarbonixL496
    1182: ["ArduPilot"],  # CBU-H7-LC-Stamp
    1156: ["ArduPilot"],  # CBU-H7-Stamp
    1177: ["ArduPilot"],  # CrazyF405
    12: ["ArduPilot"],  # crazyflie2
    1158: ["CSKY"],  # CSKY405
    7000: ["ArduPilot"],  # CUAV-7-Nano
    1009: ["ArduPilot"],  # CUAV-Nora
    9: ["ArduPilot", "Hex/ProfiCNC"],  # CUAV-Pixhack-v3, CubeBlack, CubeGreen-solo, CubePurple, CubeSolo, fmuv2, fmuv3, mRoX21, Pixhawk1, Pixhawk1-1M, skyviper-f412-rev1, skyviper-journey, skyviper-v2450
    1010: ["ArduPilot"],  # CUAV-X7
    50: ["ArduPilot", "Holybro"],  # CUAVv5, CUAVv5Nano, fmuv5, PH4-mini, Pix32v5, Pixhawk4, Swan-K1, TBS-Colibri-F7
    1003: ["Hex/ProfiCNC"],  # CubeBlack+
    1079: ["ArduPilot"],  # CubeNode
    140: ["Hex/ProfiCNC"],  # CubeOrange
    1033: ["CubePilot"],  # CubeOrange-joey
    1063: ["CubePilot"],  # CubeOrangePlus
    1069: ["CubePilot"],  # CubeRedPrimary
    1409: ["CubePilot"],  # CubeRedPrimary-PPPGW
    1070: ["ArduPilot"],  # CubeRedSecondary, CubeRedSecondary-IO
    120: ["Hex/ProfiCNC"],  # CubeYellow
    1061: ["ArduPilot"],  # DevEBoxH7v2
    5800: ["ArduPilot"],  # DroneerF405
    13: ["ArduPilot"],  # DrotekP3Pro
    139: ["Holybro", "ArduPilot"],  # Durandal, NucleoH743, NucleoH755
    135: ["ArduPilot"],  # F35Lightning
    20: ["Swift-Flyer"],  # F4BY
    1530: ["Swift-Flyer"],  # F4BY_F427
    1067: ["ArduPilot"],  # FlyingMoonF407
    1068: ["ArduPilot"],  # FlyingMoonF427
    1112: ["ArduPilot"],  # FlyingMoonH743
    1180: ["ArduPilot"],  # FlywooF405HD-AIOv2
    1137: ["ArduPilot"],  # FlywooF405Pro
    1099: ["ArduPilot"],  # FlywooF405S-AIO
    1027: ["ArduPilot"],  # FlywooF745
    1042: ["ArduPilot"],  # FlywooF745Nano
    1181: ["ArduPilot"],  # FlywooH743Pro
    1157: ["ArduPilot"],  # FoxeerF405v2
    1089: ["ArduPilot"],  # FoxeerH743v1
    1028: ["ArduPilot"],  # FreeflyRTK
    1502: ["ArduPilot"],  # GEPRC_TAKER_H743
    1501: ["ArduPilot"],  # GEPRCF745BTHD
    146: ["ArduPilot"],  # H757I_EVAL, H757I_EVAL_intf
    1119: ["ArduPilot"],  # HEEWING-F405, HEEWING-F405v2
    1043: ["ArduPilot"],  # Here4AP, Here4FC
    5401: ["ArduPilot"],  # HolybroF4_PMU
    1173: ["ArduPilot"],  # IFLIGHT_2RAW_H7
    1084: ["ArduPilot"],  # JFB100
    1110: ["Japan Aviation Electronics Industry Ltd."],  # JFB110
    1081: ["ArduPilot"],  # JHEM_JHEF405
    1059: ["ArduPilot"],  # JHEMCU-GSF405A, JHEMCU-GSF405A-RX2
    1411: ["ArduPilot"],  # JHEMCU-H743HD
    122: ["ArduPilot"],  # KakuteF4
    5406: ["ArduPilot"],  # KakuteF4-Wing
    1030: ["ArduPilot"],  # KakuteF4Mini
    123: ["ArduPilot"],  # KakuteF7
    145: ["ArduPilot"],  # KakuteF7Mini
    1048: ["ArduPilot"],  # KakuteH7, KakuteH7v2
    1105: ["ArduPilot"],  # KakuteH7-Wing
    1058: ["ArduPilot"],  # KakuteH7Mini, KakuteH7Mini-Nand
    1315: ["ArduPilot"],  # kha_eth
    1422: ["ArduPilot"],  # LongBowF405WING
    4500: ["ArduPilot"],  # LumenierLUXF765-NDAA
    11: ["ArduPilot"],  # luminousbee4, Pixracer
    1029: ["ArduPilot"],  # luminousbee5
    1038: ["ArduPilot"],  # MambaF405-2022
    1019: ["ArduPilot"],  # MambaF405v2
    1073: ["ArduPilot"],  # MambaH743v4
    125: ["ArduPilot"],  # MatekF405, MatekF405-STD
    1014: ["ArduPilot"],  # MatekF405-CAN
    1054: ["ArduPilot"],  # MatekF405-TE
    127: ["ArduPilot"],  # MatekF405-Wing
    143: ["ArduPilot"],  # MatekF765-SE, MatekF765-Wing
    1170: ["ArduPilot"],  # MatekG474
    1013: ["ArduPilot"],  # MatekH743
    1149: ["ArduPilot"],  # MatekH7A3
    188: ["ArduPilot"],  # MazzyStarDrone
    6101: ["ArduPilot"],  # MFE_POS3_CAN
    2000: ["ArduPilot"],  # MFT-SEMA100
    1161: ["ArduPilot"],  # MicoAir405Mini
    1150: ["ArduPilot"],  # MicoAir405v2
    1166: ["ArduPilot"],  # MicoAir743
    1176: ["ArduPilot"],  # MicoAir743-AIO
    1179: ["ArduPilot"],  # MicoAir743v2
    88: ["ArduPilot"],  # mindpx-v2
    3: ["ArduPilot"],  # mini-pix
    41775: ["ArduPilot"],  # modalai_fc-v1
    1041: ["ArduPilot"],  # mRo-M10095
    1098: ["ArduPilot"],  # mRoCANPWM-M10126
    1022: ["mRo"],  # mRoControlZeroClassic
    141: ["mRo"],  # mRoControlZeroF7
    1023: ["mRo"],  # mRoControlZeroH7
    1024: ["mRo"],  # mRoControlZeroOEMH7, mRoCZeroOEMH7
    1015: ["mRo"],  # mRoNexus
    1017: ["mRo"],  # mRoPixracerPro
    136: ["mRo"],  # mRoX21-777
    1222: ["ArduPilot"],  # MUPilot
    1040: ["ArduPilot"],  # Nucleo-G491
    1051: ["ArduPilot"],  # Nucleo-L476
    1047: ["ArduPilot"],  # Nucleo-L496
    1159: ["ArduPilot"],  # NxtPX4v2
    1002: ["ArduPilot"],  # omnibusf4
    131: ["ArduPilot"],  # omnibusf4pro, omnibusf4pro-one
    137: ["ArduPilot"],  # omnibusf4v6
    121: ["ArduPilot"],  # OMNIBUSF7V2
    133: ["ArduPilot"],  # OmnibusNanoV6
    1155: ["ArduPilot"],  # OrqaF405Pro
    1032: ["Horizon31"],  # PixC4-Jetson
    1090: ["Dheeran labs"],  # PixFlamingo
    1131: ["Dheeran labs"],  # PixFlamingo-F767
    51: ["Holybro"],  # Pixhawk5X
    56: ["Holybro"],  # Pixhawk6C
    53: ["ArduPilot"],  # Pixhawk6X
    1408: ["ArduPilot"],  # Pixhawk6X-PPPGW
    1140: ["ArduPilot"],  # PixPilot-C3
    1096: ["ArduPilot"],  # PixPilot-V3
    1083: ["ArduPilot"],  # PixPilot-V6
    1160: ["ArduPilot"],  # PixPilot-V6PRO
    1076: ["ArduPilot"],  # PixSurveyA1
    1107: ["ArduPilot"],  # PixSurveyA1-IND
    1097: ["ArduPilot"],  # PixSurveyA2
    1065: ["Qiotek"],  # QioTekAdeptF407
    1021: ["ArduPilot"],  # QioTekZealotF427
    1036: ["ArduPilot"],  # QioTekZealotH743
    1008: ["ArduPilot"],  # R9Pilot
    1410: ["ArduPilot"],  # RadiolinkPIX6
    1118: ["ArduPilot"],  # RADIX2HD
    1074: ["ArduPilot"],  # ReaperF745
    124: ["ArduPilot"],  # revo-mini, revo-mini-i2c, revo-mini-sd
    1102: ["ArduPilot"],  # rFCU
    1103: ["ArduPilot"],  # rGNSS
    1111: ["ArduPilot"],  # SDMODELH7V1
    1167: ["ArduPilot"],  # SDMODELH7V2
    1052: ["ArduPilot"],  # Sierra-F405
    1055: ["ArduPilot"],  # Sierra-F412
    1034: ["ArduPilot"],  # Sierra-F9P
    1050: ["ArduPilot"],  # Sierra-L431
    1095: ["Sierra Aerospace"],  # Sierra-PrecisionPoint
    5302: ["ArduPilot"],  # Sierra-TrueNavIC
    1091: ["ArduPilot"],  # Sierra-TrueNavPro
    5301: ["ArduPilot"],  # Sierra-TrueNavPro-G4
    1093: ["ArduPilot"],  # Sierra-TrueNorth
    1094: ["ArduPilot"],  # Sierra-TrueSpeed
    1123: ["ArduPilot"],  # SIYI_N7
    2714: ["SkySakura"],  # SkySakuraH743
    1075: ["ArduPilot"],  # SkystarsH7HD
    130: ["ArduPilot"],  # sparky2
    134: ["ArduPilot"],  # speedybeef4
    5271: ["ArduPilot"],  # SpeedyBeeF405AIO
    1135: ["ArduPilot"],  # SpeedyBeeF405Mini
    1106: ["ArduPilot"],  # SpeedyBeeF405WING
    1082: ["ArduPilot"],  # speedybeef4v3
    1136: ["ArduPilot"],  # speedybeef4v4
    1060: ["ArduPilot"],  # SPRacingH7
    1108: ["ArduPilot"],  # SPRacingH7RF
    1500: ["ArduPilot"],  # StellarF4
    1504: ["ArduPilot"],  # StellarF4V2
    1503: ["ArduPilot"],  # StellarH7V2
    1011: ["ArduPilot"],  # SuccexF4
    2005: ["ArduPilot"],  # SULILGH7-P1-P2
    6001: ["ArduPilot"],  # sw-boom-f407
    6002: ["ArduPilot"],  # sw-nav-f405
    6000: ["ArduPilot"],  # sw-spar-f407
    5250: ["ArduPilot"],  # TBS_LUCID_H7
    5251: ["ArduPilot"],  # TBS_LUCID_PRO
    212: ["ThePeach"],  # thepeach-k1
    213: ["ThePeach"],  # thepeach-r1
    1138: ["ArduPilot"],  # TMotorH743
    1151: ["Laser Navigation"],  # VRBrain-v51
    1152: ["Laser Navigation"],  # VRBrain-v52
    1154: ["Laser Navigation"],  # VRBrain-v54
    1910: ["Laser Navigation"],  # VRCore-v10
    1351: ["Laser Navigation"],  # VRUBrain-v51
    7100: ["ArduPilot"],  # VUAV-V7pro
    1174: ["ArduPilot"],  # X-MAV-AP-H743v2
    1113: ["ArduPilot"],  # YJUAV_A6
    1127: ["ArduPilot"],  # YJUAV_A6SE
    1141: ["ArduPilot"],  # YJUAV_A6SE_H743
    1144: ["ArduPilot"],  # YJUAV_A6Ultra
    5600: ["ArduPilot"],  # ZeroOneX6, ZeroOneX6_Air
}

# Maps 16-bit APJ board ID to MCU series for 246 supported boards
APJ_BOARD_ID_MCU_SERIES_DICT: dict[int, list[str]] = {
    1124: ["STM32H7xx"],  # 3DRControlZeroG
    1115: ["STM32F4xx"],  # ACNS-CM4Pilot
    1116: ["STM32F4xx"],  # ACNS-F405AIO
    1109: ["STM32L431"],  # AeroFox-GNSS_F9P
    7110: ["STM32H7xx"],  # AEROFOX-H7
    1080: ["STM32L431"],  # AeroFox-PMU
    2024: ["STM32H7xx"],  # AET-H743-Basic
    128: ["STM32F4xx"],  # airbotf4
    55: ["STM32F7xx"],  # AIRLink
    5200: ["STM32H7xx"],  # Airvolute-DCS2
    1146: ["STM32H7xx"],  # AnyleafH7
    5210: ["STM32H7xx"],  # Aocoda-RC-H743Dual
    1134: ["CKS32F4xx"],  # AR-F407SmartBat
    83: ["STM32F4xx"],  # ARK_CANNODE
    59: ["STM32H7xx"],  # ARK_FPV
    57: ["STM32H7xx"],  # ARKV6X
    1078: ["STM32F4xx"],  # AtomRCF405NAVI
    1026: ["STM32F7xx"],  # BeastF7
    1057: ["STM32F7xx"],  # BeastF7v2
    1025: ["STM32H7xx"],  # BeastH7
    1056: ["STM32H7xx"],  # BeastH7v2
    1125: ["STM32F4xx"],  # BETAFPV-F405
    1044: ["STM32F4xx"],  # BirdCANdy
    1164: ["STM32F7xx"],  # BlitzF745
    1117: ["STM32F7xx"],  # BlitzF745AIO
    1162: ["STM32H7xx"],  # BlitzH743Pro
    1163: ["STM32F7xx"],  # BlitzMiniF745
    1168: ["STM32H7xx"],  # BlitzWingH743
    1148: ["STM32H7xx"],  # BotBloxDroneNet
    1085: ["STM32F4xx"],  # C-RTK2-HP
    1064: ["STM32F4xx"],  # CarbonixF405
    1053: ["STM32L496"],  # CarbonixL496
    1182: ["STM32H7xx"],  # CBU-H7-LC-Stamp
    1156: ["STM32H7xx"],  # CBU-H7-Stamp
    1177: ["STM32F4xx"],  # CrazyF405
    12: ["STM32F4xx"],  # crazyflie2
    1158: ["STM32F4xx"],  # CSKY405
    7000: ["STM32H7xx"],  # CUAV-7-Nano
    1009: ["STM32H7xx"],  # CUAV-Nora
    9: ["STM32F4xx"],  # CUAV-Pixhack-v3, CubeBlack, CubeGreen-solo, CubePurple, CubeSolo, fmuv2, fmuv3, mRoX21, Pixhawk1, Pixhawk1-1M, skyviper-f412-rev1, skyviper-journey, skyviper-v2450
    1010: ["STM32H7xx"],  # CUAV-X7
    50: ["STM32F7xx"],  # CUAVv5, CUAVv5Nano, fmuv5, PH4-mini, Pix32v5, Pixhawk4, Swan-K1, TBS-Colibri-F7
    1003: ["STM32F4xx"],  # CubeBlack+
    1079: ["STM32H7xx"],  # CubeNode
    140: ["STM32H7xx"],  # CubeOrange
    1033: ["STM32H7xx"],  # CubeOrange-joey
    1063: ["STM32H7xx"],  # CubeOrangePlus
    1069: ["STM32H7xx"],  # CubeRedPrimary
    1409: ["STM32H7xx"],  # CubeRedPrimary-PPPGW
    1070: ["STM32H7xx"],  # CubeRedSecondary, CubeRedSecondary-IO
    120: ["STM32F7xx"],  # CubeYellow
    1061: ["STM32H7xx"],  # DevEBoxH7v2
    5800: ["STM32F4xx"],  # DroneerF405
    13: ["STM32F4xx"],  # DrotekP3Pro
    139: ["STM32H7xx"],  # Durandal, NucleoH743, NucleoH755
    135: ["STM32F4xx"],  # F35Lightning
    20: ["STM32F4xx"],  # F4BY
    1530: ["STM32F4xx"],  # F4BY_F427
    1067: ["STM32F4xx"],  # FlyingMoonF407
    1068: ["STM32F4xx"],  # FlyingMoonF427
    1112: ["STM32H7xx"],  # FlyingMoonH743
    1180: ["STM32F4xx"],  # FlywooF405HD-AIOv2
    1137: ["STM32F4xx"],  # FlywooF405Pro
    1099: ["STM32F4xx"],  # FlywooF405S-AIO
    1027: ["STM32F7xx"],  # FlywooF745
    1042: ["STM32F7xx"],  # FlywooF745Nano
    1181: ["STM32H7xx"],  # FlywooH743Pro
    1157: ["STM32F4xx"],  # FoxeerF405v2
    1089: ["STM32H7xx"],  # FoxeerH743v1
    1028: ["STM32F7xx"],  # FreeflyRTK
    1502: ["STM32H7xx"],  # GEPRC_TAKER_H743
    1501: ["STM32F7xx"],  # GEPRCF745BTHD
    146: ["STM32H7xx"],  # H757I_EVAL, H757I_EVAL_intf
    1119: ["STM32F4xx"],  # HEEWING-F405, HEEWING-F405v2
    1043: ["STM32H7xx"],  # Here4AP, Here4FC
    5401: ["STM32F4xx"],  # HolybroF4_PMU
    1173: ["STM32H7xx"],  # IFLIGHT_2RAW_H7
    1084: ["STM32F7xx"],  # JFB100
    1110: ["STM32H7xx"],  # JFB110
    1081: ["STM32F4xx"],  # JHEM_JHEF405
    1059: ["STM32F4xx"],  # JHEMCU-GSF405A, JHEMCU-GSF405A-RX2
    1411: ["STM32H7xx"],  # JHEMCU-H743HD
    122: ["STM32F4xx"],  # KakuteF4
    5406: ["STM32F4xx"],  # KakuteF4-Wing
    1030: ["STM32F4xx"],  # KakuteF4Mini
    123: ["STM32F7xx"],  # KakuteF7
    145: ["STM32F7xx"],  # KakuteF7Mini
    1048: ["STM32H7xx"],  # KakuteH7, KakuteH7v2
    1105: ["STM32H7xx"],  # KakuteH7-Wing
    1058: ["STM32H7xx"],  # KakuteH7Mini, KakuteH7Mini-Nand
    1315: ["STM32H7xx"],  # kha_eth
    1422: ["STM32F4xx"],  # LongBowF405WING
    4500: ["STM32F7xx"],  # LumenierLUXF765-NDAA
    11: ["STM32F4xx"],  # luminousbee4, Pixracer
    1029: ["STM32H7xx"],  # luminousbee5
    1038: ["STM32F4xx"],  # MambaF405-2022
    1019: ["STM32F4xx"],  # MambaF405v2
    1073: ["STM32H7xx"],  # MambaH743v4
    125: ["STM32F4xx"],  # MatekF405, MatekF405-STD
    1014: ["STM32F4xx"],  # MatekF405-CAN
    1054: ["STM32F4xx"],  # MatekF405-TE
    127: ["STM32F4xx"],  # MatekF405-Wing
    143: ["STM32F7xx"],  # MatekF765-SE, MatekF765-Wing
    1170: ["STM32G474"],  # MatekG474
    1013: ["STM32H7xx"],  # MatekH743
    1149: ["STM32H7xx"],  # MatekH7A3
    188: ["STM32F7xx"],  # MazzyStarDrone
    6101: ["STM32F4xx"],  # MFE_POS3_CAN
    2000: ["STM32H7xx"],  # MFT-SEMA100
    1161: ["STM32F4xx"],  # MicoAir405Mini
    1150: ["STM32F4xx"],  # MicoAir405v2
    1166: ["STM32H7xx"],  # MicoAir743
    1176: ["STM32H7xx"],  # MicoAir743-AIO
    1179: ["STM32H7xx"],  # MicoAir743v2
    88: ["STM32F4xx"],  # mindpx-v2
    3: ["STM32F4xx"],  # mini-pix
    41775: ["STM32F7xx"],  # modalai_fc-v1
    1041: ["STM32G4xx"],  # mRo-M10095
    1098: ["STM32G4xx"],  # mRoCANPWM-M10126
    1022: ["STM32H7xx"],  # mRoControlZeroClassic
    141: ["STM32F7xx"],  # mRoControlZeroF7
    1023: ["STM32H7xx"],  # mRoControlZeroH7
    1024: ["STM32H7xx"],  # mRoControlZeroOEMH7, mRoCZeroOEMH7
    1015: ["STM32H7xx"],  # mRoNexus
    1017: ["STM32H7xx"],  # mRoPixracerPro
    136: ["STM32F7xx"],  # mRoX21-777
    1222: ["STM32F7xx"],  # MUPilot
    1040: ["STM32G491"],  # Nucleo-G491
    1051: ["STM32L476"],  # Nucleo-L476
    1047: ["STM32L496"],  # Nucleo-L496
    1159: ["STM32H7xx"],  # NxtPX4v2
    1002: ["STM32F4xx"],  # omnibusf4
    131: ["STM32F4xx"],  # omnibusf4pro, omnibusf4pro-one
    137: ["STM32F4xx"],  # omnibusf4v6
    121: ["STM32F7xx"],  # OMNIBUSF7V2
    133: ["STM32F4xx"],  # OmnibusNanoV6
    1155: ["STM32F4xx"],  # OrqaF405Pro
    1032: ["STM32H7xx"],  # PixC4-Jetson
    1090: ["STM32L4xx"],  # PixFlamingo
    1131: ["STM32F7xx"],  # PixFlamingo-F767
    51: ["STM32F7xx"],  # Pixhawk5X
    56: ["STM32H7xx"],  # Pixhawk6C
    53: ["STM32H7xx"],  # Pixhawk6X
    1408: ["STM32H7xx"],  # Pixhawk6X-PPPGW
    1140: ["STM32F4xx"],  # PixPilot-C3
    1096: ["STM32F4xx"],  # PixPilot-V3
    1083: ["STM32H7xx"],  # PixPilot-V6
    1160: ["STM32H7xx"],  # PixPilot-V6PRO
    1076: ["STM32F4xx"],  # PixSurveyA1
    1107: ["STM32F4xx"],  # PixSurveyA1-IND
    1097: ["STM32H7xx"],  # PixSurveyA2
    1065: ["CKS32F4xx"],  # QioTekAdeptF407
    1021: ["STM32F4xx"],  # QioTekZealotF427
    1036: ["STM32H7xx"],  # QioTekZealotH743
    1008: ["STM32F7xx"],  # R9Pilot
    1410: ["STM32F7xx"],  # RadiolinkPIX6
    1118: ["STM32H7xx"],  # RADIX2HD
    1074: ["STM32F7xx"],  # ReaperF745
    124: ["STM32F4xx"],  # revo-mini, revo-mini-i2c, revo-mini-sd
    1102: ["STM32H7xx"],  # rFCU
    1103: ["STM32G4xx"],  # rGNSS
    1111: ["STM32H7xx"],  # SDMODELH7V1
    1167: ["STM32H7xx"],  # SDMODELH7V2
    1052: ["STM32F4xx"],  # Sierra-F405
    1055: ["STM32F4xx"],  # Sierra-F412
    1034: ["STM32F4xx"],  # Sierra-F9P
    1050: ["STM32L431"],  # Sierra-L431
    1095: ["STM32F4xx"],  # Sierra-PrecisionPoint
    5302: ["STM32L431"],  # Sierra-TrueNavIC
    1091: ["STM32L431"],  # Sierra-TrueNavPro
    5301: ["STM32G4xx"],  # Sierra-TrueNavPro-G4
    1093: ["STM32L431"],  # Sierra-TrueNorth
    1094: ["STM32L431"],  # Sierra-TrueSpeed
    1123: ["STM32H7xx"],  # SIYI_N7
    2714: ["STM32H7xx"],  # SkySakuraH743
    1075: ["STM32H7xx"],  # SkystarsH7HD
    130: ["STM32F4xx"],  # sparky2
    134: ["STM32F4xx"],  # speedybeef4
    5271: ["STM32F4xx"],  # SpeedyBeeF405AIO
    1135: ["STM32F4xx"],  # SpeedyBeeF405Mini
    1106: ["STM32F4xx"],  # SpeedyBeeF405WING
    1082: ["STM32F4xx"],  # speedybeef4v3
    1136: ["STM32F4xx"],  # speedybeef4v4
    1060: ["STM32H7xx"],  # SPRacingH7
    1108: ["STM32H7xx"],  # SPRacingH7RF
    1500: ["STM32F4xx"],  # StellarF4
    1504: ["STM32F4xx"],  # StellarF4V2
    1503: ["STM32H7xx"],  # StellarH7V2
    1011: ["STM32F4xx"],  # SuccexF4
    2005: ["STM32H7xx"],  # SULILGH7-P1-P2
    6001: ["STM32F4xx"],  # sw-boom-f407
    6002: ["STM32F4xx"],  # sw-nav-f405
    6000: ["STM32F4xx"],  # sw-spar-f407
    5250: ["STM32H7xx"],  # TBS_LUCID_H7
    5251: ["STM32F4xx"],  # TBS_LUCID_PRO
    212: ["STM32F4xx"],  # thepeach-k1
    213: ["STM32F4xx"],  # thepeach-r1
    1138: ["STM32H7xx"],  # TMotorH743
    1151: ["STM32F4xx"],  # VRBrain-v51
    1152: ["STM32F4xx"],  # VRBrain-v52
    1154: ["STM32F4xx"],  # VRBrain-v54
    1910: ["STM32F4xx"],  # VRCore-v10
    1351: ["STM32F4xx"],  # VRUBrain-v51
    7100: ["STM32H7xx"],  # VUAV-V7pro
    1174: ["STM32H7xx"],  # X-MAV-AP-H743v2
    1113: ["STM32H7xx"],  # YJUAV_A6
    1127: ["STM32H7xx"],  # YJUAV_A6SE
    1141: ["STM32H7xx"],  # YJUAV_A6SE_H743
    1144: ["STM32H7xx"],  # YJUAV_A6Ultra
    5600: ["STM32H7xx"],  # ZeroOneX6, ZeroOneX6_Air
}

# Maps MCU series to 16-bit APJ board ID for 246 supported boards
MCU_SERIES_APJ_BOARD_ID_DICT: dict[str, list[int]] = {
    "CKS32F4xx": [1134, 1065],  # AR-F407SmartBat, QioTekAdeptF407
    "STM32F4xx": [1115, 1116, 128, 83, 1078, 1125, 1044, 1085, 1064, 1177, 12, 1158, 9, 1003, 5800, 13, 135, 20, 1530, 1067, 1068, 1180, 1137, 1099, 1157, 1119, 5401, 1081, 1059, 122, 5406, 1030, 1422, 11, 1038, 1019, 125, 1014, 1054, 127, 6101, 1161, 1150, 88, 3, 1002, 131, 137, 133, 1155, 1140, 1096, 1076, 1107, 1021, 124, 1052, 1055, 1034, 1095, 130, 134, 5271, 1135, 1106, 1082, 1136, 1500, 1504, 1011, 6001, 6002, 6000, 5251, 212, 213, 1151, 1152, 1154, 1910, 1351],  # ACNS-CM4Pilot, ACNS-F405AIO, airbotf4, ARK_CANNODE, AtomRCF405NAVI, BETAFPV-F405, BirdCANdy, C-RTK2-HP, CarbonixF405, CrazyF405, crazyflie2, CSKY405, CUAV-Pixhack-v3, CubeBlack, CubeGreen-solo, CubePurple, CubeSolo, fmuv2, fmuv3, mRoX21, Pixhawk1, Pixhawk1-1M, skyviper-f412-rev1, skyviper-journey, skyviper-v2450, CubeBlack+, DroneerF405, DrotekP3Pro, F35Lightning, F4BY, F4BY_F427, FlyingMoonF407, FlyingMoonF427, FlywooF405HD-AIOv2, FlywooF405Pro, FlywooF405S-AIO, FoxeerF405v2, HEEWING-F405, HEEWING-F405v2, HolybroF4_PMU, JHEM_JHEF405, JHEMCU-GSF405A, JHEMCU-GSF405A-RX2, KakuteF4, KakuteF4-Wing, KakuteF4Mini, LongBowF405WING, luminousbee4, Pixracer, MambaF405-2022, MambaF405v2, MatekF405, MatekF405-STD, MatekF405-CAN, MatekF405-TE, MatekF405-Wing, MFE_POS3_CAN, MicoAir405Mini, MicoAir405v2, mindpx-v2, mini-pix, omnibusf4, omnibusf4pro, omnibusf4pro-one, omnibusf4v6, OmnibusNanoV6, OrqaF405Pro, PixPilot-C3, PixPilot-V3, PixSurveyA1, PixSurveyA1-IND, QioTekZealotF427, revo-mini, revo-mini-i2c, revo-mini-sd, Sierra-F405, Sierra-F412, Sierra-F9P, Sierra-PrecisionPoint, sparky2, speedybeef4, SpeedyBeeF405AIO, SpeedyBeeF405Mini, SpeedyBeeF405WING, speedybeef4v3, speedybeef4v4, StellarF4, StellarF4V2, SuccexF4, sw-boom-f407, sw-nav-f405, sw-spar-f407, TBS_LUCID_PRO, thepeach-k1, thepeach-r1, VRBrain-v51, VRBrain-v52, VRBrain-v54, VRCore-v10, VRUBrain-v51
    "STM32F7xx": [55, 1026, 1057, 1164, 1117, 1163, 50, 120, 1027, 1042, 1028, 1501, 1084, 123, 145, 4500, 143, 188, 41775, 141, 136, 1222, 121, 1131, 51, 1008, 1410, 1074],  # AIRLink, BeastF7, BeastF7v2, BlitzF745, BlitzF745AIO, BlitzMiniF745, CUAVv5, CUAVv5Nano, fmuv5, PH4-mini, Pix32v5, Pixhawk4, Swan-K1, TBS-Colibri-F7, CubeYellow, FlywooF745, FlywooF745Nano, FreeflyRTK, GEPRCF745BTHD, JFB100, KakuteF7, KakuteF7Mini, LumenierLUXF765-NDAA, MatekF765-SE, MatekF765-Wing, MazzyStarDrone, modalai_fc-v1, mRoControlZeroF7, mRoX21-777, MUPilot, OMNIBUSF7V2, PixFlamingo-F767, Pixhawk5X, R9Pilot, RadiolinkPIX6, ReaperF745
    "STM32G474": [1170],  # MatekG474
    "STM32G491": [1040],  # Nucleo-G491
    "STM32G4xx": [1041, 1098, 1103, 5301],  # mRo-M10095, mRoCANPWM-M10126, rGNSS, Sierra-TrueNavPro-G4
    "STM32H7xx": [1124, 7110, 2024, 5200, 1146, 5210, 59, 57, 1025, 1056, 1162, 1168, 1148, 1182, 1156, 7000, 1009, 1010, 1079, 140, 1033, 1063, 1069, 1409, 1070, 1061, 139, 1112, 1181, 1089, 1502, 146, 1043, 1173, 1110, 1411, 1048, 1105, 1058, 1315, 1029, 1073, 1013, 1149, 2000, 1166, 1176, 1179, 1022, 1023, 1024, 1015, 1017, 1159, 1032, 56, 53, 1408, 1083, 1160, 1097, 1036, 1118, 1102, 1111, 1167, 1123, 2714, 1075, 1060, 1108, 1503, 2005, 5250, 1138, 7100, 1174, 1113, 1127, 1141, 1144, 5600],  # 3DRControlZeroG, AEROFOX-H7, AET-H743-Basic, Airvolute-DCS2, AnyleafH7, Aocoda-RC-H743Dual, ARK_FPV, ARKV6X, BeastH7, BeastH7v2, BlitzH743Pro, BlitzWingH743, BotBloxDroneNet, CBU-H7-LC-Stamp, CBU-H7-Stamp, CUAV-7-Nano, CUAV-Nora, CUAV-X7, CubeNode, CubeOrange, CubeOrange-joey, CubeOrangePlus, CubeRedPrimary, CubeRedPrimary-PPPGW, CubeRedSecondary, CubeRedSecondary-IO, DevEBoxH7v2, Durandal, NucleoH743, NucleoH755, FlyingMoonH743, FlywooH743Pro, FoxeerH743v1, GEPRC_TAKER_H743, H757I_EVAL, H757I_EVAL_intf, Here4AP, Here4FC, IFLIGHT_2RAW_H7, JFB110, JHEMCU-H743HD, KakuteH7, KakuteH7v2, KakuteH7-Wing, KakuteH7Mini, KakuteH7Mini-Nand, kha_eth, luminousbee5, MambaH743v4, MatekH743, MatekH7A3, MFT-SEMA100, MicoAir743, MicoAir743-AIO, MicoAir743v2, mRoControlZeroClassic, mRoControlZeroH7, mRoControlZeroOEMH7, mRoCZeroOEMH7, mRoNexus, mRoPixracerPro, NxtPX4v2, PixC4-Jetson, Pixhawk6C, Pixhawk6X, Pixhawk6X-PPPGW, PixPilot-V6, PixPilot-V6PRO, PixSurveyA2, QioTekZealotH743, RADIX2HD, rFCU, SDMODELH7V1, SDMODELH7V2, SIYI_N7, SkySakuraH743, SkystarsH7HD, SPRacingH7, SPRacingH7RF, StellarH7V2, SULILGH7-P1-P2, TBS_LUCID_H7, TMotorH743, VUAV-V7pro, X-MAV-AP-H743v2, YJUAV_A6, YJUAV_A6SE, YJUAV_A6SE_H743, YJUAV_A6Ultra, ZeroOneX6, ZeroOneX6_Air
    "STM32L431": [1109, 1080, 1050, 5302, 1091, 1093, 1094],  # AeroFox-GNSS_F9P, AeroFox-PMU, Sierra-L431, Sierra-TrueNavIC, Sierra-TrueNavPro, Sierra-TrueNorth, Sierra-TrueSpeed
    "STM32L476": [1051],  # Nucleo-L476
    "STM32L496": [1053, 1047],  # CarbonixL496, Nucleo-L496
    "STM32L4xx": [1090],  # PixFlamingo
}



"""
TKinter base classes reused in multiple parts of the code.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

# https://wiki.tcl-lang.org/page/Changing+Widget+Colors

import tkinter as tk

# from logging import debug as logging_debug
# from logging import info as logging_info
from logging import error as logging_error
from tkinter import ttk
from typing import Optional, Union

from PIL import Image, ImageTk

from ardupilot_methodic_configurator import _
from ardupilot_methodic_configurator.backend_filesystem import LocalFilesystem


class BaseWindow:
    """
    A base class for creating windows in the ArduPilot Methodic Configurator application.

    This class provides a foundation for creating windows in the application, including setting up the
    root window, applying a theme, and configuring the application icon. It also includes methods for
    creating a progress window and centering a window on its parent.
    """

    def __init__(self, root_tk: Optional[tk.Tk] = None) -> None:
        self.root: Union[tk.Toplevel, tk.Tk]
        if root_tk:
            self.root = tk.Toplevel(root_tk)
        else:
            self.root = tk.Tk()
            # Set the application icon for the window and all child windows
            # https://pythonassets.com/posts/window-icon-in-tk-tkinter/
            self.root.iconphoto(True, tk.PhotoImage(file=LocalFilesystem.application_icon_filepath()))  # noqa: FBT003

        # Set the theme to 'alt'
        style = ttk.Style()
        style.theme_use("alt")
        style.configure("Bold.TLabel", font=("TkDefaultFont", 10, "bold"))

        self.main_frame = ttk.Frame(self.root)
        self.main_frame.pack(expand=True, fill=tk.BOTH)

    @staticmethod
    def center_window(window: Union[tk.Toplevel, tk.Tk], parent: Union[tk.Toplevel, tk.Tk]) -> None:
        """
        Center a window on its parent window.

        Args:
            window (tk.Toplevel|tk.Tk): The window to center.
            parent (tk.Toplevel|tk.Tk): The parent window.

        """
        window.update_idletasks()
        parent_width = parent.winfo_width()
        parent_height = parent.winfo_height()
        window_width = window.winfo_width()
        window_height = window.winfo_height()
        x = parent.winfo_x() + (parent_width // 2) - (window_width // 2)
        y = parent.winfo_y() + (parent_height // 2) - (window_height // 2)
        window.geometry(f"+{x}+{y}")

    @staticmethod
    def put_image_in_label(parent: ttk.Frame, filepath: str, image_height: int = 40) -> ttk.Label:
        # Load the image and scale it down to image_height pixels in height
        image = Image.open(filepath)
        width, height = image.size
        aspect_ratio = width / height
        new_width = int(image_height * aspect_ratio)
        resized_image = image.resize((new_width, image_height))

        # Convert the image to a format that can be used by Tkinter
        try:
            photo = ImageTk.PhotoImage(resized_image)

            # Create a label with the resized image
            image_label = ttk.Label(parent, image=photo)
            # Keep a reference to the image to prevent it from being garbage collected
            image_label.image = photo  # type: ignore[attr-defined]
        except TypeError as e:
            logging_error(_("Error loading %s image from file: %s"), filepath, e)
            image_label = ttk.Label(parent)
        return image_label



"""
Battery cell voltages management.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

from math import nan

battery_cell_voltages = {
    "LiIon": {
        "absolute_max": 4.2,
        "absolute_min": 2.5,
        "recommended_max": 4.1,
        "recommended_low": 3.1,
        "recommended_crit": 2.8,
    },
    "LiIonSS": {
        "absolute_max": 4.2,
        "absolute_min": 2.4,
        "recommended_max": 4.2,
        "recommended_low": 3.0,
        "recommended_crit": 2.7,
    },
    "LiIonSSHV": {
        "absolute_max": 4.45,
        "absolute_min": 2.4,
        "recommended_max": 4.45,
        "recommended_low": 3.0,
        "recommended_crit": 2.7,
    },
    "Lipo": {
        "absolute_max": 4.2,
        "absolute_min": 3.0,
        "recommended_max": 4.2,
        "recommended_low": 3.6,
        "recommended_crit": 3.3,
    },
    "LipoHV": {
        "absolute_max": 4.35,
        "absolute_min": 3.0,
        "recommended_max": 4.35,
        "recommended_low": 3.6,
        "recommended_crit": 3.3,
    },
    "LipoHVSS": {
        "absolute_max": 4.2,
        "absolute_min": 2.9,
        "recommended_max": 4.2,
        "recommended_low": 3.5,
        "recommended_crit": 3.2,
    },
    # Add more chemistries as needed
}


class BatteryCell:
    """
    This class provides methods to work with battery cell voltages for different chemistries.

    It includes methods to get the tuple of chemistries, limit voltages based on chemistry type,
    and get recommended voltages for a given chemistry.
    """

    @staticmethod
    def chemistries() -> tuple[str, ...]:
        return tuple(battery_cell_voltages.keys())

    @staticmethod
    def limit_max_voltage(chemistry: str) -> float:
        max_abs_max = max(chemistry["absolute_max"] for chemistry in battery_cell_voltages.values())
        if chemistry not in battery_cell_voltages:
            return max_abs_max
        return battery_cell_voltages[chemistry].get("absolute_max", max_abs_max)

    @staticmethod
    def limit_min_voltage(chemistry: str) -> float:
        min_abs_min = min(chemistry["absolute_min"] for chemistry in battery_cell_voltages.values())
        if chemistry not in battery_cell_voltages:
            return min_abs_min
        return battery_cell_voltages[chemistry].get("absolute_min", min_abs_min)

    @staticmethod
    def recommended_max_voltage(chemistry: str) -> float:
        if chemistry not in battery_cell_voltages:
            return nan
        return battery_cell_voltages[chemistry].get("recommended_max", 4.2)

    @staticmethod
    def recommended_low_voltage(chemistry: str) -> float:
        if chemistry not in battery_cell_voltages:
            return nan
        return battery_cell_voltages[chemistry].get("recommended_low", 3.6)

    @staticmethod
    def recommended_crit_voltage(chemistry: str) -> float:
        if chemistry not in battery_cell_voltages:
            return nan
        return battery_cell_voltages[chemistry].get("recommended_crit", 3.3)



#!/usr/bin/python3

"""
Tests for the argcomplete command line completions for the user-facing scripts.

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

from argparse import ArgumentParser

from ardupilot_methodic_configurator.__main__ import create_argument_parser as main_parser
from ardupilot_methodic_configurator.annotate_params import create_argument_parser as annotate_parser
from ardupilot_methodic_configurator.backend_mavftp import create_argument_parser as mavftp_parser
from ardupilot_methodic_configurator.extract_param_defaults import create_argument_parser as extract_parser
from ardupilot_methodic_configurator.param_pid_adjustment_update import create_argument_parser as pid_parser


def test_main_parser_type() -> None:
    parser = main_parser()
    assert isinstance(parser, ArgumentParser)
    assert hasattr(parser, "_actions")
    assert any(hasattr(action, "completer") for action in parser._actions)  # pylint: disable=protected-access


def test_extract_defaults_parser_type() -> None:
    parser = extract_parser()
    assert isinstance(parser, ArgumentParser)
    assert hasattr(parser, "_actions")
    assert any(hasattr(action, "completer") for action in parser._actions)  # pylint: disable=protected-access


def test_annotate_params_parser_type() -> None:
    parser = annotate_parser()
    assert isinstance(parser, ArgumentParser)
    assert hasattr(parser, "_actions")
    assert any(hasattr(action, "completer") for action in parser._actions)  # pylint: disable=protected-access


def test_pid_adjustment_parser_type() -> None:
    parser = pid_parser()
    assert isinstance(parser, ArgumentParser)
    assert hasattr(parser, "_actions")
    assert any(hasattr(action, "completer") for action in parser._actions)  # pylint: disable=protected-access


def test_mavftp_parser_type() -> None:
    parser = mavftp_parser()
    assert isinstance(parser, ArgumentParser)
    assert hasattr(parser, "_actions")
    assert any(hasattr(action, "completer") for action in parser._actions)  # pylint: disable=protected-access



#!/usr/bin/env python3

"""
Tests for the backend_filesystem_configuration_steps.py file.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import unittest
from unittest.mock import mock_open, patch

from ardupilot_methodic_configurator.backend_filesystem_configuration_steps import ConfigurationSteps


class TestConfigurationSteps(unittest.TestCase):
    """ConfigurationSteps test class."""

    def setUp(self) -> None:
        self.config_steps = ConfigurationSteps("vehicle_dir", "vehicle_type")

    @patch("builtins.open", new_callable=mock_open, read_data='{"steps": {"test_file": {"forced_parameters": {}}}}')
    @patch("os.path.join")
    @patch("os.path.dirname")
    @patch("os.path.abspath")
    def test_re_init(
        self,
        mock_abspath: unittest.mock.Mock,
        mock_dirname: unittest.mock.Mock,
        mock_join: unittest.mock.Mock,
        mock_open2: unittest.mock.Mock,
    ) -> None:
        mock_abspath.return_value = "abs_path"
        mock_dirname.return_value = "dir_name"
        mock_join.side_effect = lambda *args: "/".join(args)
        self.config_steps.re_init("vehicle_dir", "vehicle_type")
        assert self.config_steps.configuration_steps
        mock_open2.assert_has_calls(
            [
                unittest.mock.call("vehicle_dir/configuration_steps_vehicle_type.json", encoding="utf-8"),
                unittest.mock.call("dir_name/configuration_steps_schema.json", encoding="utf-8"),
            ],
            any_order=True,
        )
        assert "test_file" in self.config_steps.configuration_steps

    def test_compute_parameters(self) -> None:
        file_info = {"forced_parameters": {"PARAM1": {"New Value": "10", "Change Reason": "Test reason"}}}
        variables: dict[str, dict] = {"doc_dict": {"PARAM1": {"values": {10: "value"}}}}
        result = self.config_steps.compute_parameters("test_file", file_info, "forced", variables)
        assert result == ""
        assert "test_file" in self.config_steps.forced_parameters
        assert "PARAM1" in self.config_steps.forced_parameters["test_file"]
        assert self.config_steps.forced_parameters["test_file"]["PARAM1"].value == 10
        assert self.config_steps.forced_parameters["test_file"]["PARAM1"].comment == "Test reason"

    def test_auto_changed_by(self) -> None:
        self.config_steps.configuration_steps = {"test_file": {"auto_changed_by": "auto_change"}}
        result = self.config_steps.auto_changed_by("test_file")
        assert result == "auto_change"

    def test_jump_possible(self) -> None:
        self.config_steps.configuration_steps = {"test_file": {"jump_possible": {"step1": "step2"}}}
        result = self.config_steps.jump_possible("test_file")
        assert result == {"step1": "step2"}

    def test_get_documentation_text_and_url(self) -> None:
        self.config_steps.configuration_steps = {
            "test_file": {"prefix_text": "Documentation text", "prefix_url": "http://example.com"}
        }
        text, url = self.config_steps.get_documentation_text_and_url("test_file", "prefix")
        assert text == "Documentation text"
        assert url == "http://example.com"

    def test_get_seq_tooltip_text(self) -> None:
        self.config_steps.configuration_steps = {"test_file": {"tooltip_key": "Tooltip text"}}
        result = self.config_steps.get_seq_tooltip_text("test_file", "tooltip_key")
        assert result == "Tooltip text"

    def test_missing_new_value(self) -> None:
        file_info = {"forced_parameters": {"PARAM1": {"Change Reason": "Test reason"}}}
        with self.assertLogs(level="ERROR") as log:
            self.config_steps._ConfigurationSteps__validate_parameters_in_configuration_steps("test_file", file_info, "forced")  # pylint: disable=protected-access
            assert any("New Value" in message for message in log.output)

    def test_missing_change_reason(self) -> None:
        file_info = {"forced_parameters": {"PARAM1": {"New Value": "10"}}}
        with self.assertLogs(level="ERROR") as log:
            self.config_steps._ConfigurationSteps__validate_parameters_in_configuration_steps("test_file", file_info, "forced")  # pylint: disable=protected-access
            assert any("Change Reason" in message for message in log.output)

    def test_compute_parameters_with_invalid_expression(self) -> None:
        file_info = {"forced_parameters": {"PARAM1": {"New Value": "invalid_expression", "Change Reason": "Test reason"}}}
        variables: dict[str, dict] = {"doc_dict": {"PARAM1": {"values": {10: "value"}}}}
        result = self.config_steps.compute_parameters("test_file", file_info, "forced", variables)
        assert "could not be computed" in result

    def test_compute_parameters_with_missing_doc_dict(self) -> None:
        file_info = {"forced_parameters": {"PARAM1": {"New Value": "10", "Change Reason": "Test reason"}}}
        variables: dict[str, dict] = {}
        result = self.config_steps.compute_parameters("test_file", file_info, "forced", variables)
        assert result == ""

    def test_compute_parameters_with_string_result(self) -> None:
        file_info = {"forced_parameters": {"PARAM1": {"New Value": "'value'", "Change Reason": "Test reason"}}}
        variables: dict[str, dict] = {"doc_dict": {"PARAM1": {"values": {10: "value"}}}}
        result = self.config_steps.compute_parameters("test_file", file_info, "forced", variables)
        assert result == ""
        assert self.config_steps.forced_parameters["test_file"]["PARAM1"].value == 10


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

"""
Component editor GUI that is not data dependent.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import tkinter as tk
from argparse import ArgumentParser, Namespace

# from logging import debug as logging_debug
from logging import basicConfig as logging_basicConfig
from logging import getLevelName as logging_getLevelName
from logging import info as logging_info
from sys import exit as sys_exit
from tkinter import messagebox, ttk
from typing import Any, Union

from ardupilot_methodic_configurator import _, __version__
from ardupilot_methodic_configurator.backend_filesystem import LocalFilesystem
from ardupilot_methodic_configurator.common_arguments import add_common_arguments
from ardupilot_methodic_configurator.frontend_tkinter_base_window import BaseWindow
from ardupilot_methodic_configurator.frontend_tkinter_component_template_manager import ComponentTemplateManager
from ardupilot_methodic_configurator.frontend_tkinter_rich_text import RichText
from ardupilot_methodic_configurator.frontend_tkinter_scroll_frame import ScrollFrame
from ardupilot_methodic_configurator.frontend_tkinter_show import show_error_message, show_tooltip
from ardupilot_methodic_configurator.frontend_tkinter_usage_popup_window import UsagePopupWindow


def argument_parser() -> Namespace:
    """
    Parses command-line arguments for the script.

    This function sets up an argument parser to handle the command-line arguments for the script.

    Returns:
    argparse.Namespace: An object containing the parsed arguments.

    """
    # pylint: disable=duplicate-code
    parser = ArgumentParser(
        description=_(
            "A GUI for editing JSON files that contain vehicle component configurations. "
            "Not to be used directly, but through the main ArduPilot methodic configurator script."
        )
    )
    parser = LocalFilesystem.add_argparse_arguments(parser)
    parser = ComponentEditorWindowBase.add_argparse_arguments(parser)
    return add_common_arguments(parser).parse_args()
    # pylint: enable=duplicate-code


class ComponentEditorWindowBase(BaseWindow):
    """
    A class for editing JSON files in the ArduPilot methodic configurator.

    This class provides a graphical user interface for editing JSON files that
    contain vehicle component configurations. It inherits from the BaseWindow
    class, which provides basic window functionality.
    """

    def __init__(self, version: str, local_filesystem: LocalFilesystem) -> None:
        super().__init__()
        self.local_filesystem = local_filesystem

        self.root.title(_("Amilcar Lucas's - ArduPilot methodic configurator ") + version + _(" - Vehicle Component Editor"))
        self.root.geometry("880x600")  # Set the window width

        self.root.protocol("WM_DELETE_WINDOW", self.on_closing)

        self.data = local_filesystem.load_vehicle_components_json_data(local_filesystem.vehicle_dir)
        if len(self.data) < 1:
            # Schedule the window to be destroyed after the mainloop has started
            self.root.after(100, self.root.destroy)  # Adjust the delay as needed
            return

        self.entry_widgets: dict[tuple, Union[ttk.Entry, ttk.Combobox]] = {}

        intro_frame = ttk.Frame(self.main_frame)
        intro_frame.pack(side=tk.TOP, fill="x", expand=False)

        style = ttk.Style()
        style.configure("bigger.TLabel", font=("TkDefaultFont", 13))
        style.configure("comb_input_invalid.TCombobox", fieldbackground="red")
        style.configure("comb_input_valid.TCombobox", fieldbackground="white")
        style.configure("entry_input_invalid.TEntry", fieldbackground="red")
        style.configure("entry_input_valid.TEntry", fieldbackground="white")
        style.configure("Optional.TLabelframe", borderwidth=2)
        style.configure("Optional.TLabelframe.Label", foreground="gray")

        explanation_text = _("Please configure properties of the vehicle components.\n")
        explanation_text += _("Labels for optional properties are displayed in gray text.\n")
        explanation_text += _("Scroll down to ensure that you do not overlook any properties.\n")
        explanation_label = ttk.Label(intro_frame, text=explanation_text, wraplength=800, justify=tk.LEFT)
        explanation_label.configure(style="bigger.TLabel")
        explanation_label.pack(side=tk.LEFT, padx=(10, 10), pady=(10, 0), anchor=tk.NW)

        # Load the vehicle image and scale it down to image_height pixels in height
        if local_filesystem.vehicle_image_exists():
            image_label = self.put_image_in_label(intro_frame, local_filesystem.vehicle_image_filepath(), 100)
            image_label.pack(side=tk.RIGHT, anchor=tk.NE, padx=(4, 4), pady=(4, 0))
            show_tooltip(image_label, _("Replace the vehicle.jpg file in the vehicle directory to change the vehicle image."))
        else:
            image_label = ttk.Label(intro_frame, text=_("Add a 'vehicle.jpg' image file to the vehicle directory."))
            image_label.pack(side=tk.RIGHT, anchor=tk.NE, padx=(4, 4), pady=(4, 0))

        self.scroll_frame = ScrollFrame(self.main_frame)
        self.scroll_frame.pack(side="top", fill="both", expand=True)

        self.update_json_data()

        save_frame = ttk.Frame(self.main_frame)
        save_frame.pack(side=tk.TOP, fill="x", expand=False)
        self.save_button = ttk.Button(
            save_frame, text=_("Save data and start configuration"), command=self.validate_and_save_component_json
        )
        show_tooltip(
            self.save_button,
            _("Save component data to the vehicle_components.json file\nand start parameter value configuration and tuning."),
        )
        self.save_button.pack(pady=7)
        if UsagePopupWindow.should_display("component_editor"):
            self.root.after(10, self.__display_component_editor_usage_instructions(self.root))  # type: ignore[arg-type]

        def update_data_callback(comp_name: str, template_data: dict) -> None:
            self.data["Components"][comp_name] = template_data

        self.template_manager = ComponentTemplateManager(
            self.root,
            self.entry_widgets,
            self.get_component_data_from_gui,
            update_data_callback,
            self.derive_initial_template_name,
            local_filesystem.save_component_to_system_templates,
        )

    @staticmethod
    def __display_component_editor_usage_instructions(parent: tk.Tk) -> None:
        usage_popup_window = BaseWindow(parent)
        style = ttk.Style()

        instructions_text = RichText(
            usage_popup_window.main_frame, wrap=tk.WORD, height=5, bd=0, background=style.lookup("TLabel", "background")
        )
        instructions_text.insert(tk.END, _("1. Describe the properties of the vehicle components in the window below.\n"))
        instructions_text.insert(tk.END, _("2. Each field has mouse-over tooltips for additional guidance.\n"))
        instructions_text.insert(tk.END, _("3. Optional fields are marked with gray text and can be left blank.\n"))
        instructions_text.insert(tk.END, _("4. Scroll to the bottom of the window to ensure all properties are edited.\n"))
        instructions_text.insert(tk.END, _("5. Press the "))
        instructions_text.insert(tk.END, _("Save data and start configuration"), "italic")
        instructions_text.insert(tk.END, _(" button only after verifying that all information is correct.\n"))
        instructions_text.config(state=tk.DISABLED)

        UsagePopupWindow.display(
            parent,
            usage_popup_window,
            _("How to use the component editor window"),
            "component_editor",
            "690x200",
            instructions_text,
        )

    def update_json_data(self) -> None:  # should be overwritten in child classes
        if "Format version" not in self.data:
            self.data["Format version"] = 1

    def _set_component_value_and_update_ui(self, path: tuple, value: str) -> None:
        data_path = self.data["Components"]
        for key in path[:-1]:
            data_path = data_path[key]
        data_path[path[-1]] = value
        entry = self.entry_widgets[path]
        entry.delete(0, tk.END)
        entry.insert(0, value)
        entry.config(state="disabled")

    def populate_frames(self) -> None:
        """Populates the ScrollFrame with widgets based on the JSON data."""
        if "Components" in self.data:
            for key, value in self.data["Components"].items():
                self._add_widget(self.scroll_frame.view_port, key, value, [])

    def _add_widget(self, parent: tk.Widget, key: str, value: Union[dict, float], path: list) -> None:
        """
        Adds a widget to the parent widget with the given key and value.

        Args:
            parent (tkinter.Widget): The parent widget to which the LabelFrame/Entry will be added.
            key (str): The key for the LabelFrame/Entry.
            value (dict|float): The value associated with the key.
            path (list): The path to the current key in the JSON data.

        """
        if isinstance(value, dict):  # JSON non-leaf elements, add LabelFrame widget
            self.__add_non_leaf_widget(parent, key, value, path)
        else:  # JSON leaf elements, add Entry widget
            self.__add_leaf_widget(parent, key, value, path)

    def __add_non_leaf_widget(self, parent: tk.Widget, key: str, value: dict, path: list) -> None:
        is_toplevel = parent == self.scroll_frame.view_port
        pady = 5 if is_toplevel else 3

        current_path = (*path, key)
        description, is_optional = self.local_filesystem.get_component_property_description(current_path)

        if is_optional:
            frame = ttk.LabelFrame(parent, text=_(key), style="Optional.TLabelframe")
        else:
            frame = ttk.LabelFrame(parent, text=_(key))

        frame.pack(
            fill=tk.X, side=tk.TOP if is_toplevel else tk.LEFT, pady=pady, padx=5, anchor=tk.NW if is_toplevel else tk.N
        )

        # Enhance tooltip for optional fields
        if is_optional and description:
            description += _("\nThis is optional and can be left blank")

        # Add tooltip based on schema description
        if description:
            show_tooltip(frame, _(description), position_below=False)

        if is_toplevel and key in self.data.get("Components", {}):
            self._add_template_controls(frame, key)

        for sub_key, sub_value in value.items():
            # recursively add child elements
            self._add_widget(frame, sub_key, sub_value, [*path, key])

    def __add_leaf_widget(self, parent: tk.Widget, key: str, value: float, path: list) -> None:
        entry_frame = ttk.Frame(parent)
        entry_frame.pack(side=tk.TOP, fill=tk.X, pady=(0, 5))

        description, is_optional = self.local_filesystem.get_component_property_description((*path, key))

        label = ttk.Label(entry_frame, text=_(key), foreground="gray" if is_optional else "black")

        label.pack(side=tk.LEFT)

        entry = self.add_entry_or_combobox(value, entry_frame, (*path, key), is_optional)
        entry.pack(side=tk.LEFT, expand=True, fill=tk.X, padx=(0, 5))

        # Store the entry widget in the entry_widgets dictionary for later retrieval
        self.entry_widgets[(*path, key)] = entry

        # Enhance tooltip for optional fields
        if is_optional and description:
            description += _("\nThis is optional and can be left blank")

        # Add tooltip based on schema description
        if description:
            show_tooltip(label, _(description))
            show_tooltip(entry, _(description))

    def _add_template_controls(self, parent_frame: ttk.LabelFrame, component_name: str) -> None:
        """Add template controls for a component."""
        self.template_manager.add_template_controls(parent_frame, component_name)

    def derive_initial_template_name(self, component_data: dict[str, Any]) -> str:  # pylint: disable=unused-argument # noqa: ARG002
        """
        Derive an initial template name from the component data.

        This is a basic implementation that can be overridden in derived classes
        to provide more specific template naming logic.
        """
        return ""

    def get_component_data_from_gui(self, component_name: str) -> dict[str, Any]:
        """Save the current component configuration as a template."""
        # Get fresh component data from the GUI elements instead of stored data
        component_data: dict[str, Any] = {}

        # Find all entry widgets belonging to this component and extract their values
        for path, entry in self.entry_widgets.items():
            if len(path) >= 1 and path[0] == component_name:
                # Get the current value from the entry widget
                value: Union[str, int, float] = entry.get()

                # Try to convert to appropriate type (int, float, or string)
                if path[-1] != "Version":
                    try:
                        value = int(value)
                    except ValueError:
                        try:
                            value = float(value)
                        except ValueError:
                            value = str(value).strip()

                # Create nested structure by navigating through the path
                current_level = component_data
                for key in path[1:-1]:  # Skip component_name and the last key
                    if key not in current_level:
                        current_level[key] = {}
                    current_level = current_level[key]

                # Set the value at the final level
                current_level[path[-1]] = value
        return component_data

    def validate_and_save_component_json(self) -> None:
        """Saves the edited JSON data back to the file."""
        confirm_message = _(
            "ArduPilot Methodic Configurator only operates correctly if all component properties are correct."
            " ArduPilot parameter values depend on the components used and their connections.\n\n"
            " Have you used the scrollbar on the right side of the window and "
            "entered the correct values for all components?"
        )
        user_confirmation = messagebox.askyesno(_("Confirm that all component properties are correct"), confirm_message)

        if not user_confirmation:
            # User chose 'No', so return and do not save data
            return

        self.save_component_json()

    def save_component_json(self) -> None:
        # User confirmed, proceed with saving the data
        for path, entry in self.entry_widgets.items():
            value: Union[str, int, float] = entry.get()
            # Navigate through the nested dictionaries using the elements of the path
            current_data = self.data["Components"]
            for key in path[:-1]:
                current_data = current_data[key]

            if path[-1] != "Version":
                try:
                    value = int(value)
                except ValueError:
                    try:
                        value = float(value)
                    except ValueError:
                        value = str(value).strip()

            # Update the value in the data dictionary
            current_data[path[-1]] = value

        # Save the updated data back to the JSON file
        failed, msg = self.local_filesystem.save_vehicle_components_json_data(self.data, self.local_filesystem.vehicle_dir)
        if failed:
            show_error_message(_("Error"), _("Failed to save data to file.") + "\n" + msg)
        else:
            logging_info(_("Vehicle component data saved successfully."))
        self.root.destroy()

    def on_closing(self) -> None:
        """Handle window closing event."""
        answer = messagebox.askyesnocancel(_("Save Changes?"), _("Do you want to save the changes before closing?"))

        if answer is None:  # Cancel was clicked
            return

        if answer:
            self.save_component_json()
        else:
            self.root.destroy()
        sys_exit(0)

    # This function will be overwritten in child classes
    def add_entry_or_combobox(
        self,
        value: float,
        entry_frame: ttk.Frame,
        _path: tuple[str, str, str],
        is_optional: bool = False,  # pylint: disable=unused-argument # noqa: ARG002
    ) -> Union[ttk.Entry, ttk.Combobox]:
        entry = ttk.Entry(entry_frame)
        entry.insert(0, str(value))
        return entry

    @staticmethod
    def add_argparse_arguments(parser: ArgumentParser) -> ArgumentParser:
        parser.add_argument(
            "--skip-component-editor",
            action="store_true",
            help=_(
                "Skip the component editor window. Only use this if all components have been configured. "
                "Default is %(default)s"
            ),
        )
        return parser


if __name__ == "__main__":
    args = argument_parser()

    logging_basicConfig(level=logging_getLevelName(args.loglevel), format="%(asctime)s - %(levelname)s - %(message)s")

    filesystem = LocalFilesystem(
        args.vehicle_dir, args.vehicle_type, "", args.allow_editing_template_files, args.save_component_to_system_templates
    )
    app = ComponentEditorWindowBase(__version__, filesystem)
    app.root.mainloop()



#!/usr/bin/env python3

"""
Tests for middleware_software_updates.py.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

from unittest.mock import Mock, patch

import pytest

from ardupilot_methodic_configurator import _
from ardupilot_methodic_configurator.middleware_software_updates import UpdateManager, format_version_info


@pytest.fixture
def mock_dialog() -> Mock:
    dialog = Mock()
    dialog.show.return_value = True
    dialog.update_progress = Mock()
    return dialog


@pytest.fixture
def update_manager() -> UpdateManager:
    return UpdateManager()


class TestUpdateManager:
    """Test cases for the UpdateManager class."""

    def test_format_version_info(self) -> None:
        result = format_version_info("1.0.0", "2.0.0", "Test changes")
        assert "1.0.0" in result
        assert "2.0.0" in result
        assert "Test changes" in result

    def test_check_and_update_same_version(self, update_manager) -> None:  # pylint: disable=redefined-outer-name
        latest_release = {"tag_name": "v1.0.0"}
        current_version = "1.0.0"

        assert not update_manager.check_and_update(latest_release, current_version)

    def test_check_and_update_key_error(self, update_manager) -> None:  # pylint: disable=redefined-outer-name
        latest_release = {}
        current_version = "1.0.0"

        with patch("ardupilot_methodic_configurator.middleware_software_updates.logging_error") as mock_logging_error:
            assert not update_manager.check_and_update(latest_release, current_version)
            mock_logging_error.assert_called_once()

    def test_check_and_update_value_error(self, update_manager) -> None:  # pylint: disable=redefined-outer-name
        latest_release = {"tag_name": "v2.0.0"}
        current_version = "1.0.0"

        with (
            patch(
                "ardupilot_methodic_configurator.middleware_software_updates.format_version_info",
                side_effect=ValueError,
            ),
            patch("ardupilot_methodic_configurator.middleware_software_updates.logging_error") as mock_logging_error,
        ):
            assert not update_manager.check_and_update(latest_release, current_version)
            mock_logging_error.assert_called_once()


def test_format_version_info_pr_removal() -> None:
    changes = "Feature [#123) Added test\nBug [#456) Fixed issue"
    result = format_version_info("1.0.0", "2.0.0", changes)
    assert "[#123)" not in result
    assert "[#456)" not in result
    assert "Added test" in result
    assert "Fixed issue" in result


def test_format_version_info_author_removal() -> None:
    changes = "Feature ([author)) Added test\nBug ([contributor)) Fixed issue"
    result = format_version_info("1.0.0", "2.0.0", changes)
    assert "([author))" not in result
    assert "([contributor))" not in result


def test_format_version_info_complex_changes() -> None:
    changes = "Feature [#123)([author)) Multiple tags\nBug [#456)([contributor)) Mixed content"
    result = format_version_info("1.0.0", "2.0.0", changes)
    assert "[#123)" not in result
    assert "[#456)" not in result
    assert "([author))" not in result
    assert "([contributor))" not in result
    assert "Multiple tags" in result
    assert "Mixed content" in result


def test_format_version_info_empty_changes() -> None:
    result = format_version_info("1.0.0", "2.0.0", "")
    assert "Current version: 1.0.0" in result
    assert "Latest version: 2.0.0" in result
    assert "Changes:" in result


def test_format_version_info_special_chars() -> None:
    changes = "Feature ([#123]) Added *special* characters\nBug ([#456]) with $symbols%"
    result = format_version_info("1.0.0", "2.0.0", changes)
    assert "*special*" in result
    assert "$symbols%" in result


def test_format_version_info_basic() -> None:
    result = format_version_info("1.0.0", "2.0.0", "Simple change")
    expected = (
        # pylint: disable=duplicate-code
        _("Current version: {_current_version}")
        + "\n"
        + _("Latest version: {_latest_release}")
        + "\n\n"
        + _("Changes:\n{changes}")
        # pylint: enable=duplicate-code
    ).format(_current_version="1.0.0", _latest_release="2.0.0", changes="Simple change")
    assert result == expected


def test_format_version_info_newlines() -> None:
    result = format_version_info("1.0.0", "2.0.0", "Change 1\nChange 2")
    expected = (
        # pylint: disable=duplicate-code
        _("Current version: {_current_version}")
        + "\n"
        + _("Latest version: {_latest_release}")
        + "\n\n"
        + _("Changes:\n{changes}")
        # pylint: enable=duplicate-code
    ).format(_current_version="1.0.0", _latest_release="2.0.0", changes="Change 1\nChange 2")
    assert result == expected


def test_format_version_info_empty() -> None:
    result = format_version_info("1.0.0", "2.0.0", "")
    assert "Current version: 1.0.0" in result
    assert "Latest version: 2.0.0" in result
    assert "Changes:" in result


def test_format_version_info_pr_references() -> None:
    changes = "Feature [#123) Test\nBug [#456) Fix"
    result = format_version_info("1.0.0", "2.0.0", changes)
    assert "[#123)" not in result
    assert "[#456)" not in result
    assert "Feature Test" in result
    assert "Bug Fix" in result


def test_format_version_info_malformed_refs() -> None:
    changes = "Feature [#123 Test\nBug (#456)] Fix"
    result = format_version_info("1.0.0", "2.0.0", changes)
    assert "Feature" in result
    assert "Bug" in result
    assert "Test" in result
    assert "Fix" in result



#!/usr/bin/env python3

"""
Tests for the StageProgressBar class.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import tkinter as tk
import unittest
from unittest.mock import patch

from ardupilot_methodic_configurator.frontend_tkinter_stage_progress import StageProgressBar


class TestStageProgressBar(unittest.TestCase):
    """Test cases for the StageProgressBar class."""

    def setUp(self) -> None:
        """Set up test fixtures."""
        self.root = tk.Tk()
        self.test_phases = {
            "Phase 1": {"start": 1, "description": "First phase"},
            "Phase 2": {"start": 10, "description": "Second phase", "optional": True},
            "Phase 3": {"start": 20},
            "Milestone": {"description": "A milestone without start"},
        }
        self.total_steps = 30
        self.progress_bar = StageProgressBar(self.root, self.test_phases, self.total_steps)

    def tearDown(self) -> None:
        """Clean up after tests."""
        self.root.destroy()

    def test_initialization(self) -> None:
        """Test proper initialization of StageProgressBar."""
        assert self.progress_bar.total_files == self.total_steps
        assert self.progress_bar.phases == self.test_phases

        # Should only create frames for phases with start positions
        assert len(self.progress_bar.phase_frames) == 3
        assert "Milestone" not in self.progress_bar.phase_frames

    def test_phase_ordering(self) -> None:
        """Test that phases are ordered correctly by start position."""
        phase_names = list(self.progress_bar.phase_frames.keys())
        assert phase_names == ["Phase 1", "Phase 2", "Phase 3"]

    def test_progress_updates(self) -> None:
        """Test progress bar updates for different file positions."""
        test_cases = [
            (5, [4, 0, 0]),  # In Phase 1 (1-9)
            (15, [9, 5, 0]),  # In Phase 2 (10-19)
            (25, [9, 10, 5]),  # In Phase 3 (20-30)
            (29, [9, 10, 9]),  # Still in Phase 3, but not beyond total
        ]

        for current_file, expected_values in test_cases:
            self.progress_bar.update_progress(current_file)
            for test_bar, expected in zip(self.progress_bar.phase_bars, expected_values):
                assert test_bar["bar"]["value"] == expected

    def test_invalid_progress_update(self) -> None:
        """Test handling of invalid progress updates."""
        with self.assertLogs(level="ERROR"):
            self.progress_bar.update_progress(0)
            self.progress_bar.update_progress(31)

    @patch("tkinter.ttk.Label.configure")
    def test_resize_handling(self, mock_configure) -> None:
        """Test window resize handling."""
        self.root.geometry("400x300")
        self.progress_bar._on_resize()  # pylint: disable=protected-access

        # Should update wraplength for all phase labels
        expected_calls = len(self.progress_bar.phase_frames)
        assert mock_configure.call_count == expected_calls

    def test_optional_phase_styling(self) -> None:
        """Test that optional phases are styled correctly."""
        optional_frame = self.progress_bar.phase_frames["Phase 2"]
        normal_frame = self.progress_bar.phase_frames["Phase 1"]

        # Find labels in frames
        optional_label = None
        normal_label = None
        for child in optional_frame.winfo_children():
            if isinstance(child, tk.ttk.Label):
                optional_label = child
        for child in normal_frame.winfo_children():
            if isinstance(child, tk.ttk.Label):
                normal_label = child

        assert str(optional_label.cget("foreground")) == "gray"
        assert str(normal_label.cget("foreground")) == "black"

    def test_phase_frame_creation(self) -> None:
        """Test that phase frames are created with correct structure."""
        test_frame = self.progress_bar.phase_frames["Phase 1"]

        # Check frame structure
        children = test_frame.winfo_children()
        progressbar = None
        label = None
        for child in children:
            if isinstance(child, tk.ttk.Progressbar):
                progressbar = child
            elif isinstance(child, tk.ttk.Label):
                label = child

        assert progressbar is not None, "Progressbar should exist"
        assert label is not None, "Label should exist"
        assert str(progressbar.cget("orient")) == "horizontal"
        assert str(progressbar.cget("mode")) == "determinate"

    def test_progress_bar_limits(self) -> None:
        """Test progress bar maximum values are set correctly."""
        for i, phase in enumerate(self.progress_bar.phase_bars):
            test_bar = phase["bar"]
            start = phase["start"]
            end = phase["end"]

            # Check maximum is correctly set to phase range
            assert test_bar.cget("maximum") == end - start, f"Phase {i + 1} maximum should be {end - start}"

    def test_label_text_splitting(self) -> None:
        """Test text splitting behavior for different label lengths."""
        test_texts = {
            "A": {"start": 1},  # Very short
            "Short": {"start": 2},  # Short
            "Medium Text": {"start": 3},  # Medium with space
            "MediumNoSpace": {"start": 4},  # Medium without space
            "Very Long Phase Name Example": {"start": 5},  # Long with spaces
        }

        test_progress = StageProgressBar(self.root, test_texts, 10)

        for phase_name, frame in test_progress.phase_frames.items():
            label = None
            for child in frame.winfo_children():
                if isinstance(child, tk.ttk.Label):
                    label = child
                    break

            assert label is not None
            text = label.cget("text")

            if len(phase_name) <= 1:
                # Very short text should have padding newline
                assert text.endswith("\n "), f"Very short text '{phase_name}' missing padding newline"
            elif len(phase_name) < 6:
                # Short text should have newline
                assert "\n" in text, f"Short text '{phase_name}' missing newline"
            elif " " in phase_name and len(phase_name) < 20:
                # Medium text with space should split at first space
                first_space_pos = phase_name.find(" ")
                expected_newline_pos = text.find("\n")
                assert expected_newline_pos == first_space_pos, f"Medium text '{phase_name}' not split at first space"


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

"""
Creates the ardupilot_methodic_configurator pip python package.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import fnmatch
import os

from setuptools import setup


# recursively find all files that match the globs and return tuples with their directory and a list of relative paths
def find_data_files(globs: list[str]) -> list[tuple[str, list[str]]]:
    data_files_path_base = "ardupilot_methodic_configurator"
    ret = []
    for dirpath, _dirnames, filenames in os.walk(data_files_path_base):
        data_files = []
        for glob in globs:
            for filename in fnmatch.filter(filenames, glob):
                relative_path = os.path.join(dirpath, filename)
                data_files.append(relative_path)
        if data_files:
            ret.append((os.path.relpath(dirpath, data_files_path_base), data_files))
    return ret


setup(
    packages=["ardupilot_methodic_configurator"],
    # this is used by bdist
    data_files=[*find_data_files(["*.param", "*.jpg", "*.json", "*.xml", "*.mo"])],
)



#!/usr/bin/env python3

"""
Data-dependent part of the component editor GUI.

This file is part of Ardupilot methodic configurator. https://github.com/ArduPilot/MethodicConfigurator

SPDX-FileCopyrightText: 2024-2025 Amilcar do Carmo Lucas <amilcar.lucas@iav.de>

SPDX-License-Identifier: GPL-3.0-or-later
"""

import tkinter as tk
from argparse import ArgumentParser, Namespace

# from logging import debug as logging_debug
# from logging import info as logging_info
from logging import basicConfig as logging_basicConfig
from logging import error as logging_error
from logging import getLevelName as logging_getLevelName
from math import log2
from tkinter import ttk
from typing import Any, Callable, Optional, Union

from ardupilot_methodic_configurator import _, __version__
from ardupilot_methodic_configurator.backend_filesystem import LocalFilesystem
from ardupilot_methodic_configurator.backend_filesystem_vehicle_components import VehicleComponents
from ardupilot_methodic_configurator.battery_cell_voltages import BatteryCell
from ardupilot_methodic_configurator.common_arguments import add_common_arguments
from ardupilot_methodic_configurator.frontend_tkinter_component_editor_base import ComponentEditorWindowBase

# from ardupilot_methodic_configurator.frontend_tkinter_show import show_tooltip
from ardupilot_methodic_configurator.frontend_tkinter_show import show_error_message


def argument_parser() -> Namespace:
    """
    Parses command-line arguments for the script.

    This function sets up an argument parser to handle the command-line arguments for the script.

    Returns:
    argparse.Namespace: An object containing the parsed arguments.

    """
    # pylint: disable=duplicate-code
    parser = ArgumentParser(
        description=_(
            "A GUI for editing JSON files that contain vehicle component configurations. "
            "Not to be used directly, but through the main ArduPilot methodic configurator script."
        )
    )
    parser = LocalFilesystem.add_argparse_arguments(parser)
    parser = ComponentEditorWindow.add_argparse_arguments(parser)
    return add_common_arguments(parser).parse_args()
    # pylint: enable=duplicate-code


class VoltageTooLowError(Exception):
    """Raised when the voltage is below the minimum limit."""


class VoltageTooHighError(Exception):
    """Raised when the voltage is above the maximum limit."""


analog_ports = ["Analog"]
serial_ports = ["SERIAL1", "SERIAL2", "SERIAL3", "SERIAL4", "SERIAL5", "SERIAL6", "SERIAL7", "SERIAL8"]
can_ports = ["CAN1", "CAN2"]
i2c_ports = ["I2C1", "I2C2", "I2C3", "I2C4"]
pwm_ports = ["Main Out", "AIO"]
rc_ports = ["RCin/SBUS"]

serial_protocols_dict: dict[str, dict[str, Any]] = {
    "-1": {"type": serial_ports, "protocol": "None", "component": None},
    "1": {"type": serial_ports, "protocol": "MAVLink1", "component": "Telemetry"},
    "2": {"type": serial_ports, "protocol": "MAVLink2", "component": "Telemetry"},
    "3": {"type": serial_ports, "protocol": "Frsky D", "component": None},
    "4": {"type": serial_ports, "protocol": "Frsky SPort", "component": None},
    "5": {"type": serial_ports, "protocol": "GPS", "component": "GNSS Receiver"},
    "7": {"type": serial_ports, "protocol": "Alexmos Gimbal Serial", "component": None},
    "8": {"type": serial_ports, "protocol": "Gimbal", "component": None},
    "9": {"type": serial_ports, "protocol": "Rangefinder", "component": None},
    "10": {"type": serial_ports, "protocol": "FrSky SPort Passthrough (OpenTX)", "component": None},
    "11": {"type": serial_ports, "protocol": "Lidar360", "component": None},
    "13": {"type": serial_ports, "protocol": "Beacon", "component": None},
    "14": {"type": serial_ports, "protocol": "Volz servo out", "component": None},
    "15": {"type": serial_ports, "protocol": "SBus servo out", "component": None},
    "16": {"type": serial_ports, "protocol": "ESC Telemetry", "component": None},
    "17": {"type": serial_ports, "protocol": "Devo Telemetry", "component": None},
    "18": {"type": serial_ports, "protocol": "OpticalFlow", "component": None},
    "19": {"type": serial_ports, "protocol": "RobotisServo", "component": None},
    "20": {"type": serial_ports, "protocol": "NMEA Output", "component": None},
    "21": {"type": serial_ports, "protocol": "WindVane", "component": None},
    "22": {"type": serial_ports, "protocol": "SLCAN", "component": None},
    "23": {"type": serial_ports, "protocol": "RCIN", "component": "RC Receiver"},
    "24": {"type": serial_ports, "protocol": "EFI Serial", "component": None},
    "25": {"type": serial_ports, "protocol": "LTM", "component": None},
    "26": {"type": serial_ports, "protocol": "RunCam", "component": None},
    "27": {"type": serial_ports, "protocol": "HottTelem", "component": None},
    "28": {"type": serial_ports, "protocol": "Scripting", "component": None},
    "29": {"type": serial_ports, "protocol": "Crossfire VTX", "component": None},
    "30": {"type": serial_ports, "protocol": "Generator", "component": None},
    "31": {"type": serial_ports, "protocol": "Winch", "component": None},
    "32": {"type": serial_ports, "protocol": "MSP", "component": None},
    "33": {"type": serial_ports, "protocol": "DJI FPV", "component": None},
    "34": {"type": serial_ports, "protocol": "AirSpeed", "component": None},
    "35": {"type": serial_ports, "protocol": "ADSB", "component": None},
    "36": {"type": serial_ports, "protocol": "AHRS", "component": None},
    "37": {"type": serial_ports, "protocol": "SmartAudio", "component": None},
    "38": {"type": serial_ports, "protocol": "FETtecOneWire", "component": "ESC"},
    "39": {"type": serial_ports, "protocol": "Torqeedo", "component": "ESC"},
    "40": {"type": serial_ports, "protocol": "AIS", "component": None},
    "41": {"type": serial_ports, "protocol": "CoDevESC", "component": "ESC"},
    "42": {"type": serial_ports, "protocol": "DisplayPort", "component": None},
    "43": {"type": serial_ports, "protocol": "MAVLink High Latency", "component": "Telemetry"},
    "44": {"type": serial_ports, "protocol": "IRC Tramp", "component": None},
    "45": {"type": serial_ports, "protocol": "DDS XRCE", "component": None},
    "46": {"type": serial_ports, "protocol": "IMUDATA", "component": None},
}


batt_monitor_connection: dict[str, dict[str, str]] = {
    "0": {"type": "None", "protocol": "Disabled"},
    "3": {"type": "Analog", "protocol": "Analog Voltage Only"},
    "4": {"type": "Analog", "protocol": "Analog Voltage and Current"},
    "5": {"type": "i2c", "protocol": "Solo"},
    "6": {"type": "i2c", "protocol": "Bebop"},
    "7": {"type": "i2c", "protocol": "SMBus-Generic"},
    "8": {"type": "can", "protocol": "DroneCAN-BatteryInfo"},
    "9": {"type": "None", "protocol": "ESC"},
    "10": {"type": "None", "protocol": "Sum Of Selected Monitors"},
    "11": {"type": "i2c", "protocol": "FuelFlow"},
    "12": {"type": "pwm", "protocol": "FuelLevelPWM"},
    "13": {"type": "i2c", "protocol": "SMBUS-SUI3"},
    "14": {"type": "i2c", "protocol": "SMBUS-SUI6"},
    "15": {"type": "i2c", "protocol": "NeoDesign"},
    "16": {"type": "i2c", "protocol": "SMBus-Maxell"},
    "17": {"type": "i2c", "protocol": "Generator-Elec"},
    "18": {"type": "i2c", "protocol": "Generator-Fuel"},
    "19": {"type": "i2c", "protocol": "Rotoye"},
    "20": {"type": "i2c", "protocol": "MPPT"},
    "21": {"type": "i2c", "protocol": "INA2XX"},
    "22": {"type": "i2c", "protocol": "LTC2946"},
    "23": {"type": "None", "protocol": "Torqeedo"},
    "24": {"type": "Analog", "protocol": "FuelLevelAnalog"},
    "25": {"type": "Analog", "protocol": "Synthetic Current and Analog Voltage"},
    "26": {"type": "spi", "protocol": "INA239_SPI"},
    "27": {"type": "i2c", "protocol": "EFI"},
    "28": {"type": "i2c", "protocol": "AD7091R5"},
    "29": {"type": "None", "protocol": "Scripting"},
}


gnss_receiver_connection: dict[str, Any] = {
    "0": {"type": None, "protocol": "None"},
    "1": {"type": "serial", "protocol": "AUTO"},
    "2": {"type": "serial", "protocol": "uBlox"},
    "5": {"type": "serial", "protocol": "NMEA"},
    "6": {"type": "serial", "protocol": "SiRF"},
    "7": {"type": "serial", "protocol": "HIL"},
    "8": {"type": "serial", "protocol": "SwiftNav"},
    "9": {"type": "can", "protocol": "DroneCAN"},
    "10": {"type": "serial", "protocol": "SBF"},
    "11": {"type": "serial", "protocol": "GSOF"},
    "13": {"type": "serial", "protocol": "ERB"},
    "14": {"type": "serial", "protocol": "MAV"},
    "15": {"type": "serial", "protocol": "NOVA"},
    "16": {"type": "serial", "protocol": "HemisphereNMEA"},
    "17": {"type": "serial", "protocol": "uBlox-MovingBaseline-Base"},
    "18": {"type": "serial", "protocol": "uBlox-MovingBaseline-Rover"},
    "19": {"type": "serial", "protocol": "MSP"},
    "20": {"type": "serial", "protocol": "AllyStar"},
    "21": {"type": "serial", "protocol": "ExternalAHRS"},
    "22": {"type": "can", "protocol": "DroneCAN-MovingBaseline-Base"},
    "23": {"type": "can", "protocol": "DroneCAN-MovingBaseline-Rover"},
    "24": {"type": "serial", "protocol": "UnicoreNMEA"},
    "25": {"type": "serial", "protocol": "UnicoreMovingBaselineNMEA"},
    "26": {"type": "serial", "protocol": "SBF-DualAntenna"},
}

mot_pwm_type_dict: dict[str, dict[str, Any]] = {
    "0": {"type": "Main Out", "protocol": "Normal", "is_dshot": False},
    "1": {"type": "Main Out", "protocol": "OneShot", "is_dshot": True},
    "2": {"type": "Main Out", "protocol": "OneShot125", "is_dshot": True},
    "3": {"type": "Main Out", "protocol": "Brushed", "is_dshot": False},
    "4": {"type": "Main Out", "protocol": "DShot150", "is_dshot": True},
    "5": {"type": "Main Out", "protocol": "DShot300", "is_dshot": True},
    "6": {"type": "Main Out", "protocol": "DShot600", "is_dshot": True},
    "7": {"type": "Main Out", "protocol": "DShot1200", "is_dshot": True},
    "8": {"type": "Main Out", "protocol": "PWMRange", "is_dshot": False},
}
rc_protocols_dict: dict[str, dict[str, str]] = {
    "0": {"type": "RCin/SBUS", "protocol": "All"},
    "1": {"type": "RCin/SBUS", "protocol": "PPM"},
    "2": {"type": "RCin/SBUS", "protocol": "IBUS"},
    "3": {"type": "RCin/SBUS", "protocol": "SBUS"},
    "4": {"type": "RCin/SBUS", "protocol": "SBUS_NI"},
    "5": {"type": "RCin/SBUS", "protocol": "DSM"},
    "6": {"type": "RCin/SBUS", "protocol": "SUMD"},
    "7": {"type": "RCin/SBUS", "protocol": "SRXL"},
    "8": {"type": "RCin/SBUS", "protocol": "SRXL2"},
    "9": {"type": "RCin/SBUS", "protocol": "CRSF"},
    "10": {"type": "RCin/SBUS", "protocol": "ST24"},
    "11": {"type": "RCin/SBUS", "protocol": "FPORT"},
    "12": {"type": "RCin/SBUS", "protocol": "FPORT2"},
    "13": {"type": "RCin/SBUS", "protocol": "FastSBUS"},
    "14": {"type": "can", "protocol": "DroneCAN"},
    "15": {"type": "RCin/SBUS", "protocol": "Ghost"},
}


class MockEvent(tk.Event):  # pylint: disable=too-few-public-methods
    """A mock event object that mirrors the structure of a real tkinter event."""

    def __init__(self, event_type: tk.EventType) -> None:
        self.type = event_type


class ComponentEditorWindow(ComponentEditorWindowBase):
    """
    This class validates the user input and handles user interactions
    for editing component configurations in the ArduPilot Methodic Configurator.
    """

    def __init__(self, version: str, local_filesystem: LocalFilesystem) -> None:
        self.serial_ports = ["SERIAL1", "SERIAL2", "SERIAL3", "SERIAL4", "SERIAL5", "SERIAL6", "SERIAL7", "SERIAL8"]
        self.can_ports = ["CAN1", "CAN2"]
        self.i2c_ports = ["I2C1", "I2C2", "I2C3", "I2C4"]
        ComponentEditorWindowBase.__init__(self, version, local_filesystem)
        # these are just here so that pygettext extracts them, they have no function
        _vehicle_components_strings = _("Flight Controller")
        _vehicle_components_strings = _("Product")
        _vehicle_components_strings = _("Manufacturer")
        _vehicle_components_strings = _("Model")
        _vehicle_components_strings = _("URL")
        _vehicle_components_strings = _("Version")
        _vehicle_components_strings = _("Firmware")
        _vehicle_components_strings = _("Type")
        _vehicle_components_strings = _("MCU Series")
        _vehicle_components_strings = _("Notes")
        _vehicle_components_strings = _("Frame")
        _vehicle_components_strings = _("Specifications")
        _vehicle_components_strings = _("TOW min Kg")
        _vehicle_components_strings = _("TOW max Kg")
        _vehicle_components_strings = _("RC Controller")
        _vehicle_components_strings = _("RC Transmitter")
        _vehicle_components_strings = _("RC Receiver")
        _vehicle_components_strings = _("FC Connection")
        _vehicle_components_strings = _("Protocol")
        _vehicle_components_strings = _("Telemetry")
        _vehicle_components_strings = _("Battery Monitor")
        _vehicle_components_strings = _("Battery")
        _vehicle_components_strings = _("Chemistry")
        _vehicle_components_strings = _("Volt per cell max")
        _vehicle_components_strings = _("Volt per cell low")
        _vehicle_components_strings = _("Volt per cell crit")
        _vehicle_components_strings = _("Number of cells")
        _vehicle_components_strings = _("Capacity mAh")
        _vehicle_components_strings = _("ESC")
        _vehicle_components_strings = _("Motors")
        _vehicle_components_strings = _("Poles")
        _vehicle_components_strings = _("Propellers")
        _vehicle_components_strings = _("Diameter_inches")
        _vehicle_components_strings = _("GNSS Receiver")

    def update_json_data(self) -> None:
        super().update_json_data()
        # To update old JSON files that do not have these new fields
        if "Components" not in self.data:
            self.data["Components"] = {}
        if "Battery" not in self.data["Components"]:
            self.data["Components"]["Battery"] = {}
        if "Specifications" not in self.data["Components"]["Battery"]:
            self.data["Components"]["Battery"]["Specifications"] = {}
        if "Chemistry" not in self.data["Components"]["Battery"]["Specifications"]:
            self.data["Components"]["Battery"]["Specifications"]["Chemistry"] = "Lipo"
        if "Capacity mAh" not in self.data["Components"]["Battery"]["Specifications"]:
            self.data["Components"]["Battery"]["Specifications"]["Capacity mAh"] = 0

        # To update old JSON files that do not have these new "Frame.Specifications.TOW * Kg" fields
        if "Frame" not in self.data["Components"]:
            self.data["Components"]["Frame"] = {}
        if "Specifications" not in self.data["Components"]["Frame"]:
            self.data["Components"]["Frame"]["Specifications"] = {}
        if "TOW min Kg" not in self.data["Components"]["Frame"]["Specifications"]:
            self.data["Components"]["Frame"]["Specifications"]["TOW min Kg"] = 1
        if "TOW max Kg" not in self.data["Components"]["Frame"]["Specifications"]:
            self.data["Components"]["Frame"]["Specifications"]["TOW max Kg"] = 1

        # Older versions used receiver instead of Receiver, rename it for consistency with other fields
        if "GNSS receiver" in self.data["Components"]:
            self.data["Components"]["GNSS Receiver"] = self.data["Components"].pop("GNSS receiver")

        self.data["Program version"] = __version__

        # To update old JSON files that do not have this new "Flight Controller.Specifications.MCU Series" field
        if "Flight Controller" not in self.data["Components"]:
            self.data["Components"]["Flight Controller"] = {}
        if "Specifications" not in self.data["Components"]["Flight Controller"]:
            self.data["Components"]["Flight Controller"] = {
                "Product": self.data["Components"]["Flight Controller"]["Product"],
                "Firmware": self.data["Components"]["Flight Controller"]["Firmware"],
                "Specifications": {"MCU Series": "Unknown"},
                "Notes": self.data["Components"]["Flight Controller"]["Notes"],
            }

    def set_vehicle_type_and_version(self, vehicle_type: str, version: str) -> None:
        self._set_component_value_and_update_ui(("Flight Controller", "Firmware", "Type"), vehicle_type)
        if version:
            self._set_component_value_and_update_ui(("Flight Controller", "Firmware", "Version"), version)

    def set_fc_manufacturer(self, manufacturer: str) -> None:
        if manufacturer and manufacturer not in (_("Unknown"), "ArduPilot"):
            self._set_component_value_and_update_ui(("Flight Controller", "Product", "Manufacturer"), manufacturer)

    def set_fc_model(self, model: str) -> None:
        if model and model not in (_("Unknown"), "MAVLink"):
            self._set_component_value_and_update_ui(("Flight Controller", "Product", "Model"), model)

    def set_mcu_series(self, mcu: str) -> None:
        if mcu:
            self._set_component_value_and_update_ui(("Flight Controller", "Specifications", "MCU Series"), mcu)

    def set_vehicle_configuration_template(self, configuration_template: str) -> None:
        self.data["Configuration template"] = configuration_template

    @staticmethod
    def reverse_key_search(
        doc: dict[str, dict[str, dict[str, float]]], param_name: str, values: list[float], fallbacks: list[int]
    ) -> list[int]:
        retv = [int(key) for key, value in doc[param_name]["values"].items() if value in values]
        if len(values) != len(fallbacks):
            logging_error(_("Length of values %u and fallbacks %u differ for %s"), len(values), len(fallbacks), param_name)
        if retv:
            return retv
        logging_error(_("No values found for %s in the metadata"), param_name)
        return fallbacks

    def __assert_dict_is_uptodate(self, doc: dict, dict_to_check: dict, doc_key: str, doc_dict: str) -> None:
        """Asserts that the given dictionary is up-to-date with the apm.pdef.xml documentation metadata."""
        if doc and doc_key in doc and doc[doc_key] and doc_dict in doc[doc_key]:
            for key, doc_protocol in doc[doc_key][doc_dict].items():
                if key in dict_to_check:
                    code_protocol = dict_to_check[key].get("protocol", None)
                    if code_protocol != doc_protocol:
                        logging_error(_("Protocol %s does not match %s in %s metadata"), code_protocol, doc_protocol, doc_key)
                else:
                    logging_error(_("Protocol %s not found in %s metadata"), doc_protocol, doc_key)

    def set_values_from_fc_parameters(self, fc_parameters: dict, doc: dict) -> None:
        self.__assert_dict_is_uptodate(doc, serial_protocols_dict, "SERIAL1_PROTOCOL", "values")
        self.__assert_dict_is_uptodate(doc, batt_monitor_connection, "BATT_MONITOR", "values")
        self.__assert_dict_is_uptodate(doc, gnss_receiver_connection, "GPS_TYPE", "values")
        self.__assert_dict_is_uptodate(doc, mot_pwm_type_dict, "MOT_PWM_TYPE", "values")
        self.__assert_dict_is_uptodate(doc, rc_protocols_dict, "RC_PROTOCOLS", "Bitmask")

        self.__set_gnss_type_and_protocol_from_fc_parameters(fc_parameters)
        esc_is_serial_controlled = self.__set_serial_type_and_protocol_from_fc_parameters(fc_parameters)
        if not esc_is_serial_controlled:
            self.__set_esc_type_and_protocol_from_fc_parameters(fc_parameters, doc)
        self.__set_battery_type_and_protocol_from_fc_parameters(fc_parameters)
        self.__set_motor_poles_from_fc_parameters(fc_parameters)

    def __set_gnss_type_and_protocol_from_fc_parameters(self, fc_parameters: dict) -> None:
        gps1_type = fc_parameters.get("GPS_TYPE", 0)
        try:
            gps1_type = int(gps1_type)
        except ValueError:
            logging_error(_("Invalid non-integer value for GPS_TYPE %f"), gps1_type)
            gps1_type = 0
        if str(gps1_type) in gnss_receiver_connection:
            gps1_connection_type = gnss_receiver_connection[str(gps1_type)].get("type")
            gps1_connection_protocol = gnss_receiver_connection[str(gps1_type)].get("protocol")
            if gps1_connection_type is None:
                self.data["Components"]["GNSS Receiver"]["FC Connection"]["Type"] = "None"
                self.data["Components"]["GNSS Receiver"]["FC Connection"]["Protocol"] = "None"
            elif gps1_connection_type == "serial":
                # GNSS connection type will be detected later in set_protocol_and_connection_from_fc_parameters()
                self.data["Components"]["GNSS Receiver"]["FC Connection"]["Protocol"] = gps1_connection_protocol
            elif gps1_connection_type == "can":
                if (
                    "CAN_D1_PROTOCOL" in fc_parameters
                    and fc_parameters["CAN_D1_PROTOCOL"] == 1
                    and "CAN_P1_DRIVER" in fc_parameters
                    and fc_parameters["CAN_P1_DRIVER"] == 1
                ):
                    self.data["Components"]["GNSS Receiver"]["FC Connection"]["Type"] = "CAN1"
                elif (
                    "CAN_D2_PROTOCOL" in fc_parameters
                    and fc_parameters["CAN_D2_PROTOCOL"] == 1
                    and "CAN_P2_DRIVER" in fc_parameters
                    and fc_parameters["CAN_P2_DRIVER"] == 2
                ):
                    self.data["Components"]["GNSS Receiver"]["FC Connection"]["Type"] = "CAN2"
                else:
                    logging_error(
                        _("Invalid CAN_Dx_PROTOCOL %s and CAN_Px_DRIVER %s for GNSS Receiver"),
                        fc_parameters.get("CAN_D1_PROTOCOL"),
                        fc_parameters.get("CAN_P1_DRIVER"),
                    )
                    self.data["Components"]["GNSS Receiver"]["FC Connection"]["Type"] = "None"
                self.data["Components"]["GNSS Receiver"]["FC Connection"]["Protocol"] = gps1_connection_protocol
            else:
                logging_error("Invalid GNSS connection type %s", gps1_connection_type)
                self.data["Components"]["GNSS Receiver"]["FC Connection"]["Type"] = "None"
        else:
            logging_error("GPS_TYPE %u not in gnss_receiver_connection", gps1_type)
            self.data["Components"]["GNSS Receiver"]["FC Connection"]["Type"] = "None"

    def __set_serial_type_and_protocol_from_fc_parameters(self, fc_parameters: dict[str, float]) -> bool:
        if "RC_PROTOCOLS" in fc_parameters:
            rc_protocols_nr = int(fc_parameters["RC_PROTOCOLS"])
            # check if rc_protocols_nr is a power of two (only one bit set)
            if rc_protocols_nr & (rc_protocols_nr - 1) == 0:
                # rc_bit is the number of the bit that is set
                rc_bit = str(int(log2(rc_protocols_nr)))
                protocol = rc_protocols_dict[rc_bit].get("protocol")
                self.data["Components"]["RC Receiver"]["FC Connection"]["Protocol"] = protocol

        rc = 1
        telem = 1
        gnss = 1
        esc = 1
        for serial in self.serial_ports:
            if serial + "_PROTOCOL" not in fc_parameters:
                continue
            try:
                serial_protocol_nr = int(fc_parameters[serial + "_PROTOCOL"])
            except ValueError:
                msg = _("Invalid non-integer value for {serial}_PROTOCOL {serial_protocol_nr}")
                logging_error(msg.format(**locals()))
                serial_protocol_nr = 0
            component = serial_protocols_dict[str(serial_protocol_nr)].get("component")
            protocol = serial_protocols_dict[str(serial_protocol_nr)].get("protocol")
            if component is None:
                continue
            if component == "RC Receiver" and rc == 1:
                self.data["Components"][component]["FC Connection"]["Type"] = serial  # only one RC supported
                rc += 1
            elif component == "Telemetry" and telem == 1:
                self.data["Components"][component]["FC Connection"]["Type"] = serial  # only one telemetry supported
                self.data["Components"][component]["FC Connection"]["Protocol"] = protocol
                telem += 1
            elif component == "GNSS Receiver" and gnss == 1:
                self.data["Components"][component]["FC Connection"]["Type"] = serial  # only one GNSS supported
                gnss += 1
            elif component == "ESC" and esc == 1:
                self.data["Components"][component]["FC Connection"]["Type"] = serial  # only one ESC supported
                self.data["Components"][component]["FC Connection"]["Protocol"] = protocol
                esc += 1

        return esc >= 2

    def __set_esc_type_and_protocol_from_fc_parameters(self, fc_parameters: dict, doc: dict) -> None:
        mot_pwm_type = fc_parameters.get("MOT_PWM_TYPE", 0)
        try:
            mot_pwm_type = int(mot_pwm_type)
        except ValueError:
            logging_error(_("Invalid non-integer value for MOT_PWM_TYPE %f"), mot_pwm_type)
            mot_pwm_type = 0
        main_out_functions = [fc_parameters.get("SERVO" + str(i) + "_FUNCTION", 0) for i in range(1, 9)]

        # if any element of main_out_functions is in [33, 34, 35, 36] then ESC is connected to main_out
        if any(servo_function in {33, 34, 35, 36} for servo_function in main_out_functions):
            self.data["Components"]["ESC"]["FC Connection"]["Type"] = "Main Out"
        else:
            self.data["Components"]["ESC"]["FC Connection"]["Type"] = "AIO"
        self.data["Components"]["ESC"]["FC Connection"]["Protocol"] = doc["MOT_PWM_TYPE"]["values"][str(mot_pwm_type)]

    def __set_battery_type_and_protocol_from_fc_parameters(self, fc_parameters: dict) -> None:
        if "BATT_MONITOR" in fc_parameters:
            batt_monitor = int(fc_parameters["BATT_MONITOR"])
            self.data["Components"]["Battery Monitor"]["FC Connection"]["Type"] = batt_monitor_connection[
                str(batt_monitor)
            ].get("type")
            self.data["Components"]["Battery Monitor"]["FC Connection"]["Protocol"] = batt_monitor_connection[
                str(batt_monitor)
            ].get("protocol")

    def __set_motor_poles_from_fc_parameters(self, fc_parameters: dict) -> None:
        if "MOT_PWM_TYPE" in fc_parameters:
            mot_pwm_type_str = str(fc_parameters["MOT_PWM_TYPE"])
            if mot_pwm_type_str in mot_pwm_type_dict and mot_pwm_type_dict[mot_pwm_type_str].get("is_dshot", False):
                if "SERVO_BLH_POLES" in fc_parameters:
                    self.data["Components"]["Motors"]["Specifications"]["Poles"] = fc_parameters["SERVO_BLH_POLES"]
            elif "SERVO_FTW_MASK" in fc_parameters and fc_parameters["SERVO_FTW_MASK"] and "SERVO_FTW_POLES" in fc_parameters:
                self.data["Components"]["Motors"]["Specifications"]["Poles"] = fc_parameters["SERVO_FTW_POLES"]

    def update_esc_protocol_combobox_entries(self, esc_connection_type: str) -> None:
        """Updates the ESC Protocol combobox entries based on the selected ESC Type."""
        if len(esc_connection_type) > 3 and esc_connection_type[:3] == "CAN":
            protocols = ["DroneCAN"]
        elif len(esc_connection_type) > 6 and esc_connection_type[:6] == "SERIAL":
            protocols = [value["protocol"] for value in serial_protocols_dict.values() if value["component"] == "ESC"]
        elif "MOT_PWM_TYPE" in self.local_filesystem.doc_dict:
            protocols = list(self.local_filesystem.doc_dict["MOT_PWM_TYPE"]["values"].values())
        elif "Q_M_PWM_TYPE" in self.local_filesystem.doc_dict:
            protocols = list(self.local_filesystem.doc_dict["Q_M_PWM_TYPE"]["values"].values())
        else:
            protocols = []

        protocol_path = ("ESC", "FC Connection", "Protocol")
        if protocol_path in self.entry_widgets:
            protocol_combobox = self.entry_widgets[protocol_path]
            protocol_combobox["values"] = protocols  # Update the combobox entries
            if protocol_combobox.get() not in protocols and isinstance(protocol_combobox, ttk.Combobox):
                protocol_combobox.set(protocols[0] if protocols else "")
            protocol_combobox.update_idletasks()  # re-draw the combobox ASAP

    def add_entry_or_combobox(  # pylint: disable=too-many-locals
        self, value: float, entry_frame: ttk.Frame, path: tuple[str, str, str], is_optional: bool = False
    ) -> Union[ttk.Entry, ttk.Combobox]:
        # Default values for comboboxes in case the apm.pdef.xml metadata is not available
        fallbacks: dict[str, tuple[str, ...]] = {
            "RC_PROTOCOLS": tuple(value["protocol"] for value in rc_protocols_dict.values()),
            "BATT_MONITOR": tuple(value["protocol"] for value in batt_monitor_connection.values()),
            "MOT_PWM_TYPE": tuple(value["protocol"] for value in mot_pwm_type_dict.values()),
            "GPS_TYPE": tuple(value["protocol"] for value in gnss_receiver_connection.values()),
        }

        def get_combobox_values(param_name: str) -> tuple[str, ...]:
            param_metadata = self.local_filesystem.doc_dict
            if param_name in param_metadata:
                if "values" in param_metadata[param_name] and param_metadata[param_name]["values"]:
                    return tuple(param_metadata[param_name]["values"].values())
                if "Bitmask" in param_metadata[param_name] and param_metadata[param_name]["Bitmask"]:
                    return tuple(param_metadata[param_name]["Bitmask"].values())
                logging_error(_("No values found for %s in the metadata"), param_name)
            if param_name in fallbacks:
                return fallbacks[param_name]
            logging_error(_("No fallback values found for %s"), param_name)
            return ()

        combobox_dict: dict[tuple[str, str, str], tuple[str, ...]] = {
            ("Flight Controller", "Firmware", "Type"): VehicleComponents.supported_vehicles(),
            ("RC Receiver", "FC Connection", "Type"): ("RCin/SBUS", *self.serial_ports, *self.can_ports),
            ("RC Receiver", "FC Connection", "Protocol"): get_combobox_values("RC_PROTOCOLS"),
            ("Telemetry", "FC Connection", "Type"): tuple(self.serial_ports + self.can_ports),
            ("Telemetry", "FC Connection", "Protocol"): ("MAVLink1", "MAVLink2", "MAVLink High Latency"),
            ("Battery Monitor", "FC Connection", "Type"): (
                "None",
                "Analog",
                "SPI",
                "PWM",
                *self.i2c_ports,
                *self.serial_ports,
                *self.can_ports,
            ),
            ("Battery Monitor", "FC Connection", "Protocol"): get_combobox_values("BATT_MONITOR"),
            ("ESC", "FC Connection", "Type"): ("Main Out", "AIO", *self.serial_ports, *self.can_ports),
            ("ESC", "FC Connection", "Protocol"): get_combobox_values("MOT_PWM_TYPE"),
            ("GNSS Receiver", "FC Connection", "Type"): ("None", *self.serial_ports, *self.can_ports),
            ("GNSS Receiver", "FC Connection", "Protocol"): get_combobox_values("GPS_TYPE"),
            ("Battery", "Specifications", "Chemistry"): BatteryCell.chemistries(),
        }
        combobox_values: Union[tuple[str, ...], None] = combobox_dict.get(path)

        # Determine foreground color based on is_optional flag
        fg_color = "gray" if is_optional else "black"

        if combobox_values:
            cb = ttk.Combobox(entry_frame, values=combobox_values, foreground=fg_color)
            cb.bind("<FocusOut>", lambda event, path=path: self.validate_combobox(event, path))  # type: ignore[misc]
            cb.bind("<KeyRelease>", lambda event, path=path: self.validate_combobox(event, path))  # type: ignore[misc]

            # Prevent mouse wheel from changing value when dropdown is not open
            def handle_mousewheel(_event: tk.Event, widget: tk.Widget = cb) -> Optional[str]:
                # Check if dropdown is open by examining the combobox's state
                if not hasattr(widget, "_dropdown_open") or not widget._dropdown_open:  # pylint: disable=protected-access,line-too-long # noqa: SLF001 # pyright: ignore[reportAttributeAccessIssue]
                    return "break"  # Prevent default behavior
                return None  # Allow default behavior when dropdown is open

            # Set flag when dropdown opens or closes
            def dropdown_opened(_event: tk.Event, widget: tk.Widget = cb) -> None:
                widget._dropdown_open = True  # type: ignore[attr-defined] # pylint: disable=protected-access # noqa: SLF001

            def dropdown_closed(_event: tk.Event, widget: tk.Widget = cb) -> None:
                widget._dropdown_open = False  # type: ignore[attr-defined] # pylint: disable=protected-access # noqa: SLF001

            # Initialize the flag
            cb._dropdown_open = False  # type: ignore[attr-defined] # pylint: disable=protected-access # noqa: SLF001

            # Bind to events for dropdown opening and closing
            cb.bind("<<ComboboxDropdown>>", dropdown_opened)
            cb.bind(
                "<FocusOut>",
                lambda e, p=path: (dropdown_closed(e), self.validate_combobox(e, p)),  # type: ignore[misc, func-returns-value]
            )
            # Bind mouse wheel events
            cb.bind("<MouseWheel>", handle_mousewheel)  # Windows mouse wheel
            cb.bind("<Button-4>", handle_mousewheel)  # Linux mouse wheel up
            cb.bind("<Button-5>", handle_mousewheel)  # Linux mouse wheel down

            if path == ("ESC", "FC Connection", "Type"):  # immediate update of ESC Protocol upon ESC Type selection
                cb.bind(
                    "<<ComboboxSelected>>",
                    lambda event: self.update_esc_protocol_combobox_entries(cb.get()),  # noqa: ARG005
                )

            cb.set(value)
            return cb

        entry = ttk.Entry(entry_frame, foreground=fg_color)
        validate_function = self.get_validate_function(entry, path)
        if validate_function:
            entry.bind("<FocusOut>", validate_function)
            entry.bind("<KeyRelease>", validate_function)
        entry.insert(0, str(value))
        return entry

    def get_validate_function(self, entry: ttk.Entry, path: tuple[str, str, str]) -> Union[Callable[[tk.Event], object], None]:
        validate_functions = {
            ("Frame", "Specifications", "TOW min Kg"): lambda event, entry=entry, path=path: self.validate_entry_limits(
                event, entry, float, (0.01, 600), "Takeoff Weight", path
            ),
            ("Frame", "Specifications", "TOW max Kg"): lambda event, entry=entry, path=path: self.validate_entry_limits(
                event, entry, float, (0.01, 600), "Takeoff Weight", path
            ),
            ("Battery", "Specifications", "Volt per cell max"): lambda event,
            entry=entry,
            path=path: self.validate_cell_voltage(event, entry, path),
            ("Battery", "Specifications", "Volt per cell low"): lambda event,
            entry=entry,
            path=path: self.validate_cell_voltage(event, entry, path),
            ("Battery", "Specifications", "Volt per cell crit"): lambda event,
            entry=entry,
            path=path: self.validate_cell_voltage(event, entry, path),
            ("Battery", "Specifications", "Number of cells"): lambda event, entry=entry, path=path: self.validate_entry_limits(
                event, entry, int, (1, 50), "Nr of cells", path
            ),
            ("Battery", "Specifications", "Capacity mAh"): lambda event, entry=entry, path=path: self.validate_entry_limits(
                event, entry, int, (100, 1000000), "mAh capacity", path
            ),
            ("Motors", "Specifications", "Poles"): lambda event, entry=entry, path=path: self.validate_entry_limits(
                event, entry, int, (3, 50), "Motor Poles", path
            ),
            ("Propellers", "Specifications", "Diameter_inches"): lambda event,
            entry=entry,
            path=path: self.validate_entry_limits(event, entry, float, (0.3, 400), "Propeller Diameter", path),
        }
        return validate_functions.get(path)

    def validate_combobox(self, event: tk.Event, path: tuple[str, ...]) -> bool:
        """Validates the value of a combobox."""
        combobox = event.widget  # Get the combobox widget that triggered the event
        value = combobox.get()  # Get the current value of the combobox
        allowed_values = combobox.cget("values")  # Get the list of allowed values

        if value not in allowed_values:
            if event.type == "10":  # FocusOut events
                _paths_str = ">".join(list(path))
                _allowed_str = ", ".join(allowed_values)
                error_msg = _("Invalid value '{value}' for {_paths_str}\nAllowed values are: {_allowed_str}")
                show_error_message(_("Error"), error_msg.format(**locals()))
            combobox.configure(style="comb_input_invalid.TCombobox")
            return False

        if path == ("ESC", "FC Connection", "Type"):
            self.update_esc_protocol_combobox_entries(value)

        combobox.configure(style="comb_input_valid.TCombobox")
        return True

    def validate_entry_limits(  # pylint: disable=too-many-arguments, too-many-positional-arguments
        self,
        event: Union[None, tk.Event],
        entry: ttk.Entry,
        data_type: type,
        limits: tuple[float, float],
        _name: str,
        path: tuple[str, str, str],
    ) -> bool:
        is_focusout_event = event and event.type == "10"
        try:
            value = entry.get()  # make sure value is defined to prevent exception in the except block
            value = data_type(value)
            if value < limits[0] or value > limits[1]:  # type: ignore[operator]
                entry.configure(style="entry_input_invalid.TEntry")
                error_msg = _("{_name} must be a {data_type.__name__} between {limits[0]} and {limits[1]}")
                raise ValueError(error_msg.format(**locals()))
        except ValueError as _e:
            if is_focusout_event:
                _paths_str = ">".join(list(path))
                error_msg = _("Invalid value '{value}' for {_paths_str}\n{_e}")
                show_error_message(_("Error"), error_msg.format(**locals()))
            return False
        entry.configure(style="entry_input_valid.TEntry")
        return True

    def validate_cell_voltage(self, event: Union[None, tk.Event], entry: ttk.Entry, path: tuple[str, str, str]) -> bool:  # pylint: disable=too-many-branches
        """Validates the value of a battery cell voltage entry."""
        chemistry_path = ("Battery", "Specifications", "Chemistry")
        if chemistry_path not in self.entry_widgets:
            show_error_message(_("Error"), _("Battery Chemistry not set. Will default to Lipo."))
            chemistry = "Lipo"
        else:
            chemistry = self.entry_widgets[chemistry_path].get()
        value = entry.get()
        is_focusout_event = event and event.type == "10"
        _path_str = ">".join(list(path))
        try:
            voltage = float(value)
            volt_limit = BatteryCell.limit_min_voltage(chemistry)
            if voltage < volt_limit:
                if is_focusout_event:
                    entry.delete(0, tk.END)
                    entry.insert(0, str(volt_limit))
                error_msg = _("is below the {chemistry} minimum limit of {volt_limit}")
                raise VoltageTooLowError(error_msg.format(**locals()))
            volt_limit = BatteryCell.limit_max_voltage(chemistry)
            if voltage > volt_limit:
                if is_focusout_event:
                    entry.delete(0, tk.END)
                    entry.insert(0, str(volt_limit))
                error_msg = _("is above the {chemistry} maximum limit of {volt_limit}")
                raise VoltageTooHighError(error_msg.format(**locals()))
        except (VoltageTooLowError, VoltageTooHighError) as _e:
            if is_focusout_event:
                error_msg = _("Invalid value '{value}' for {_path_str}\n{_e}")
                show_error_message(_("Error"), error_msg.format(**locals()))
            else:
                entry.configure(style="entry_input_invalid.TEntry")
                return False
        except ValueError as _e:
            if is_focusout_event:
                error_msg = _("Invalid value '{value}' for {_path_str}\n{_e}\nWill be set to the recommended value.")
                show_error_message(_("Error"), error_msg.format(**locals()))
                entry.delete(0, tk.END)
                if path[-1] == "Volt per cell max":
                    entry.insert(0, str(BatteryCell.recommended_max_voltage(chemistry)))
                elif path[-1] == "Volt per cell low":
                    entry.insert(0, str(BatteryCell.recommended_low_voltage(chemistry)))
                elif path[-1] == "Volt per cell crit":
                    entry.insert(0, str(BatteryCell.recommended_crit_voltage(chemistry)))
                else:
                    entry.insert(0, "3.8")
            else:
                entry.configure(style="entry_input_invalid.TEntry")
                return False
        entry.configure(style="entry_input_valid.TEntry")
        return True

    def validate_and_save_component_json(self) -> None:
        if self.validate_data():
            super().validate_and_save_component_json()

    def validate_data(self) -> bool:  # pylint: disable=too-many-branches
        invalid_values = False
        duplicated_connections = False
        fc_serial_connection: dict[str, str] = {}

        for path, entry in self.entry_widgets.items():
            value = entry.get()

            _path_str = ">".join(list(path))
            if isinstance(entry, ttk.Combobox):
                if path == ("ESC", "FC Connection", "Type"):
                    self.update_esc_protocol_combobox_entries(value)
                if value not in entry.cget("values"):
                    _values_str = ", ".join(entry.cget("values"))
                    error_msg = _("Invalid value '{value}' for {_path_str}\nAllowed values are: {_values_str}")
                    show_error_message(_("Error"), error_msg.format(**locals()))
                    entry.configure(style="comb_input_invalid.TCombobox")
                    invalid_values = True
                    continue
                if "FC Connection" in path and "Type" in path:
                    if value in fc_serial_connection and value not in {"CAN1", "CAN2", "I2C1", "I2C2", "I2C3", "I2C4", "None"}:
                        if path[0] in {"Telemetry", "RC Receiver"} and fc_serial_connection[value] in {
                            "Telemetry",
                            "RC Receiver",
                        }:
                            entry.configure(style="comb_input_valid.TCombobox")
                            continue  # Allow telemetry and RC Receiver connections using the same SERIAL port
                        if (
                            self.data["Components"]["Battery Monitor"]["FC Connection"]["Protocol"] == "ESC"
                            and path[0] in {"Battery Monitor", "ESC"}
                            and fc_serial_connection[value] in {"Battery Monitor", "ESC"}
                        ):
                            entry.configure(style="comb_input_valid.TCombobox")
                            continue  # Allow 'Battery Monitor' and 'ESC' connections using the same SERIAL port
                        error_msg = _("Duplicate FC connection type '{value}' for {_path_str}")
                        show_error_message(_("Error"), error_msg.format(**locals()))
                        entry.configure(style="comb_input_invalid.TCombobox")
                        duplicated_connections = True
                        continue
                    fc_serial_connection[value] = path[0]
                entry.configure(style="comb_input_valid.TCombobox")

            validate_function = self.get_validate_function(entry, path)
            if validate_function:
                mock_focus_out_event = MockEvent(tk.EventType.FocusOut)
                if not validate_function(mock_focus_out_event):
                    invalid_values = True
            if path in {
                ("Battery", "Specifications", "Volt per cell max"),
                ("Battery", "Specifications", "Volt per cell low"),
                ("Battery", "Specifications", "Volt per cell crit"),
            } and not self.validate_cell_voltage(None, entry, path):
                invalid_values = True
            if (
                path == ("Battery", "Specifications", "Volt per cell low")
                and value >= self.entry_widgets[("Battery", "Specifications", "Volt per cell max")].get()
            ):
                show_error_message(_("Error"), _("Battery Cell Low voltage must be lower than max voltage"))
                entry.configure(style="entry_input_invalid.TEntry")
                invalid_values = True
            if (
                path == ("Battery", "Specifications", "Volt per cell crit")
                and value >= self.entry_widgets[("Battery", "Specifications", "Volt per cell low")].get()
            ):
                show_error_message(_("Error"), _("Battery Cell Crit voltage must be lower than low voltage"))
                entry.configure(style="entry_input_invalid.TEntry")
                invalid_values = True

        return not (invalid_values or duplicated_connections)

    def derive_initial_template_name(self, component_data: dict[str, Any]) -> str:
        """Derive an initial template name from the component data."""
        initial_template_name: str = ""
        product_data = component_data.get("Product")
        if product_data:
            manufacturer = product_data.get("Manufacturer", "")
            model = product_data.get("Model", "")
            initial_template_name = manufacturer + " " + model
        return initial_template_name


# pylint: disable=duplicate-code
if __name__ == "__main__":
    args = argument_parser()

    logging_basicConfig(level=logging_getLevelName(args.loglevel), format="%(asctime)s - %(levelname)s - %(message)s")

    filesystem = LocalFilesystem(
        args.vehicle_dir, args.vehicle_type, "", args.allow_editing_template_files, args.save_component_to_system_templates
    )
    app = ComponentEditorWindow(__version__, filesystem)
    app.root.mainloop()
# pylint: enable=duplicate-code
